import collections
import dataclasses
import logging
from datetime import datetime
from typing import Dict, Iterable, List

from dateutil import parser
from pydantic.fields import Field
from pydantic.main import BaseModel
from sqlalchemy import create_engine
from sqlalchemy.engine import Engine

import datahub.emitter.mce_builder as builder
from datahub.configuration.source_common import EnvConfigMixin
from datahub.configuration.time_window_config import get_time_bucket
from datahub.ingestion.api.decorators import (
    SourceCapability,
    SupportStatus,
    capability,
    config_class,
    platform_name,
    support_status,
)
from datahub.ingestion.api.source import Source, SourceReport
from datahub.ingestion.api.workunit import MetadataWorkUnit
from datahub.ingestion.source.sql.clickhouse import ClickHouseConfig
from datahub.ingestion.source.usage.usage_common import (
    BaseUsageConfig,
    GenericAggregatedDataset,
)

logger = logging.getLogger(__name__)

clickhouse_datetime_format = "%Y-%m-%d %H:%M:%S"

clickhouse_usage_sql_comment = """\
SELECT user                                                                       AS username
     , query
     , substring(full_table_name, 1, position(full_table_name, '.') - 1)          AS database
     , substring(full_table_name, position(full_table_name, '.') + 1)             AS table
     , full_table_name
     , arrayMap(x -> substr(x, length(full_table_name) + 2),
                arrayFilter(x -> startsWith(x, full_table_name || '.'), columns)) AS columns
     , query_start_time                                                           AS starttime
     , event_time                                                                 AS endtime
  FROM {query_log_table}
 ARRAY JOIN tables AS full_table_name
 WHERE is_initial_query
   AND type = 'QueryFinish'
   AND query_kind = 'Select'
   AND full_table_name NOT LIKE 'system.%%'
   AND full_table_name NOT LIKE '_table_function.%%'
   AND table NOT LIKE '`.inner%%'
   AND event_time >= '{start_time}'
   AND event_time < '{end_time}'
 ORDER BY event_time DESC"""

ClickHouseTableRef = str
AggregatedDataset = GenericAggregatedDataset[ClickHouseTableRef]


class ClickHouseJoinedAccessEvent(BaseModel):
    username: str = None  # type:ignore
    query: str = None  # type: ignore
    database: str = None  # type:ignore
    table: str = None  # type:ignore
    columns: List[str]
    starttime: datetime
    endtime: datetime


class ClickHouseUsageConfig(ClickHouseConfig, BaseUsageConfig, EnvConfigMixin):
    email_domain: str = Field(description="")
    options: dict = Field(default={}, description="")
    query_log_table: str = Field(default="system.query_log", exclude=True)

    def get_sql_alchemy_url(self):
        return super().get_sql_alchemy_url()


@platform_name("ClickHouse")
@config_class(ClickHouseUsageConfig)
@support_status(SupportStatus.CERTIFIED)
@capability(SourceCapability.DELETION_DETECTION, "Enabled via stateful ingestion")
@capability(SourceCapability.DATA_PROFILING, "Optionally enabled via configuration")
@dataclasses.dataclass
class ClickHouseUsageSource(Source):
    """
    This plugin has the below functionalities -
    1. For a specific dataset this plugin ingests the following statistics -
       1. top n queries.
       2. top users.
       3. usage of each column in the dataset.
    2. Aggregation of these statistics into buckets, by day or hour granularity.

    Usage information is computed by querying the system.query_log table. In case you have a cluster or need to apply additional transformation/filters you can create a view and put to the `query_log_table` setting.

    :::note

    This source only does usage statistics. To get the tables, views, and schemas in your ClickHouse warehouse, ingest using the `clickhouse` source described above.

    :::

    """

    config: ClickHouseUsageConfig
    report: SourceReport = dataclasses.field(default_factory=SourceReport)

    @classmethod
    def create(cls, config_dict, ctx):
        config = ClickHouseUsageConfig.parse_obj(config_dict)
        return cls(ctx, config)

    def get_workunits_internal(self) -> Iterable[MetadataWorkUnit]:
        """Gets ClickHouse usage stats as work units"""
        access_events = self._get_clickhouse_history()
        # If the query results is empty, we don't want to proceed
        if not access_events:
            return []

        joined_access_event = self._get_joined_access_event(access_events)
        aggregated_info = self._aggregate_access_events(joined_access_event)

        for time_bucket in aggregated_info.values():
            for aggregate in time_bucket.values():
                yield self._make_usage_stat(aggregate)

    def _make_usage_query(self) -> str:
        return clickhouse_usage_sql_comment.format(
            query_log_table=self.config.query_log_table,
            start_time=self.config.start_time.strftime(clickhouse_datetime_format),
            end_time=self.config.end_time.strftime(clickhouse_datetime_format),
        )

    def _make_sql_engine(self) -> Engine:
        url = self.config.get_sql_alchemy_url()
        logger.debug(f"sql_alchemy_url = {url}")
        engine = create_engine(url, **self.config.options)
        return engine

    def _get_clickhouse_history(self):
        query = self._make_usage_query()
        engine = self._make_sql_engine()
        results = engine.execute(query)
        events = []
        for row in results:
            event_dict = row._asdict()

            # stripping extra spaces caused by above _asdict() conversion
            for k, v in event_dict.items():
                if isinstance(v, str):
                    event_dict[k] = v.strip()

            if not self.config.database_pattern.allowed(
                event_dict.get("database")
            ) or not (
                self.config.table_pattern.allowed(event_dict.get("full_table_name"))
                or self.config.view_pattern.allowed(event_dict.get("full_table_name"))
            ):
                logger.debug(
                    f"Dropping usage event for {event_dict.get('full_table_name')}"
                )
                continue

            if event_dict.get("starttime", None):
                event_dict["starttime"] = event_dict.get("starttime").__str__()
            if event_dict.get("endtime", None):
                event_dict["endtime"] = event_dict.get("endtime").__str__()
            # when the http protocol is used, the columns field is returned as a string
            if isinstance(event_dict.get("columns"), str):
                event_dict["columns"] = (
                    event_dict.get("columns").replace("'", "").strip("][").split(",")
                )

            logger.debug(f"event_dict: {event_dict}")
            events.append(event_dict)

        if events:
            return events

        # SQL results can be empty. If results is empty, the SQL connection closes.
        # Then, we don't want to proceed ingestion.
        logging.info("SQL Result is empty")
        return None

    def _convert_str_to_datetime(self, v):
        if isinstance(v, str):
            isodate = parser.parse(v)  # compatible with Python 3.6+
            return isodate.strftime(clickhouse_datetime_format)

    def _get_joined_access_event(self, events):
        joined_access_events = []
        for event_dict in events:
            event_dict["starttime"] = self._convert_str_to_datetime(
                event_dict.get("starttime")
            )
            event_dict["endtime"] = self._convert_str_to_datetime(
                event_dict.get("endtime")
            )

            if not (event_dict.get("database", None) and event_dict.get("table", None)):
                logging.info("An access event parameter(s) is missing. Skipping ....")
                continue

            if not event_dict.get("username") or event_dict["username"] == "":
                logging.info("The username parameter is missing. Skipping ....")
                continue

            joined_access_events.append(ClickHouseJoinedAccessEvent(**event_dict))

        return joined_access_events

    def _aggregate_access_events(
        self, events: List[ClickHouseJoinedAccessEvent]
    ) -> Dict[datetime, Dict[ClickHouseTableRef, AggregatedDataset]]:
        datasets: Dict[
            datetime, Dict[ClickHouseTableRef, AggregatedDataset]
        ] = collections.defaultdict(dict)

        for event in events:
            floored_ts = get_time_bucket(event.starttime, self.config.bucket_duration)

            resource = (
                f'{self.config.platform_instance+"." if self.config.platform_instance else ""}'
                f"{event.database}.{event.table}"
            )

            agg_bucket = datasets[floored_ts].setdefault(
                resource,
                AggregatedDataset(bucket_start_time=floored_ts, resource=resource),
            )

            # current limitation in user stats UI, we need to provide email to show users
            user_email = f"{event.username if event.username else 'unknown'}"
            if "@" not in user_email:
                user_email += f"@{self.config.email_domain}"
            logger.info(f"user_email: {user_email}")
            agg_bucket.add_read_entry(
                user_email,
                event.query,
                event.columns,
            )
        return datasets

    def _make_usage_stat(self, agg: AggregatedDataset) -> MetadataWorkUnit:
        return agg.make_usage_workunit(
            self.config.bucket_duration,
            lambda resource: builder.make_dataset_urn(
                "clickhouse", resource, self.config.env
            ),
            self.config.top_n_queries,
            self.config.format_sql_queries,
            self.config.include_top_n_queries,
            self.config.queries_character_limit,
        )

    def get_report(self) -> SourceReport:
        return self.report



import logging
from datetime import datetime
from typing import Callable, Iterable, List, Optional

from pydantic import BaseModel

from datahub.emitter.mce_builder import (
    make_assertion_urn,
    make_data_platform_urn,
    make_dataplatform_instance_urn,
)
from datahub.emitter.mcp import MetadataChangeProposalWrapper
from datahub.ingestion.api.workunit import MetadataWorkUnit
from datahub.ingestion.source.snowflake.snowflake_config import SnowflakeV2Config
from datahub.ingestion.source.snowflake.snowflake_query import SnowflakeQuery
from datahub.ingestion.source.snowflake.snowflake_report import SnowflakeV2Report
from datahub.ingestion.source.snowflake.snowflake_utils import (
    SnowflakeCommonMixin,
    SnowflakeConnectionMixin,
    SnowflakeQueryMixin,
)
from datahub.metadata.com.linkedin.pegasus2avro.assertion import (
    AssertionResult,
    AssertionResultType,
    AssertionRunEvent,
    AssertionRunStatus,
)
from datahub.metadata.com.linkedin.pegasus2avro.common import DataPlatformInstance
from datahub.utilities.time import datetime_to_ts_millis

logger: logging.Logger = logging.getLogger(__name__)


class DataQualityMonitoringResult(BaseModel):
    MEASUREMENT_TIME: datetime
    METRIC_NAME: str
    TABLE_NAME: str
    TABLE_SCHEMA: str
    TABLE_DATABASE: str
    VALUE: int


class SnowflakeAssertionsHandler(
    SnowflakeCommonMixin, SnowflakeQueryMixin, SnowflakeConnectionMixin
):
    def __init__(
        self,
        config: SnowflakeV2Config,
        report: SnowflakeV2Report,
        dataset_urn_builder: Callable[[str], str],
    ) -> None:
        self.config = config
        self.report = report
        self.logger = logger
        self.dataset_urn_builder = dataset_urn_builder
        self.connection = None
        self._urns_processed: List[str] = []

    def get_assertion_workunits(
        self, discovered_datasets: List[str]
    ) -> Iterable[MetadataWorkUnit]:

        self.connection = self.create_connection()
        if self.connection is None:
            return

        cur = self.query(
            SnowflakeQuery.dmf_assertion_results(
                datetime_to_ts_millis(self.config.start_time),
                datetime_to_ts_millis(self.config.end_time),
            )
        )
        for db_row in cur:
            mcp = self._process_result_row(db_row, discovered_datasets)
            if mcp:
                yield mcp.as_workunit(is_primary_source=False)

                if mcp.entityUrn and mcp.entityUrn not in self._urns_processed:
                    self._urns_processed.append(mcp.entityUrn)
                    yield self._gen_platform_instance_wu(mcp.entityUrn)

    def _gen_platform_instance_wu(self, urn: str) -> MetadataWorkUnit:

        # Construct a MetadataChangeProposalWrapper object for assertion platform
        return MetadataChangeProposalWrapper(
            entityUrn=urn,
            aspect=DataPlatformInstance(
                platform=make_data_platform_urn(self.platform),
                instance=(
                    make_dataplatform_instance_urn(
                        self.platform, self.config.platform_instance
                    )
                    if self.config.platform_instance
                    else None
                ),
            ),
        ).as_workunit(is_primary_source=False)

    def _process_result_row(
        self, result_row: dict, discovered_datasets: List[str]
    ) -> Optional[MetadataChangeProposalWrapper]:
        try:
            result = DataQualityMonitoringResult.parse_obj(result_row)
            assertion_guid = result.METRIC_NAME.split("__")[-1].lower()
            status = bool(result.VALUE)  # 1 if PASS, 0 if FAIL
            assertee = self.get_dataset_identifier(
                result.TABLE_NAME, result.TABLE_SCHEMA, result.TABLE_DATABASE
            )
            if assertee in discovered_datasets:
                return MetadataChangeProposalWrapper(
                    entityUrn=make_assertion_urn(assertion_guid),
                    aspect=AssertionRunEvent(
                        timestampMillis=datetime_to_ts_millis(result.MEASUREMENT_TIME),
                        runId=result.MEASUREMENT_TIME.strftime("%Y-%m-%dT%H:%M:%SZ"),
                        asserteeUrn=self.dataset_urn_builder(assertee),
                        status=AssertionRunStatus.COMPLETE,
                        assertionUrn=make_assertion_urn(assertion_guid),
                        result=AssertionResult(
                            type=(
                                AssertionResultType.SUCCESS
                                if status
                                else AssertionResultType.FAILURE
                            )
                        ),
                    ),
                )
        except Exception as e:
            self.report.report_warning("assertion-result-parse-failure", str(e))
        return None



import logging
from datetime import datetime
from functools import lru_cache
from typing import Iterable, List, Optional

from databricks.sdk.service.catalog import ColumnTypeName, DataSourceFormat
from databricks.sql.types import Row
from sqlalchemy import create_engine, inspect
from sqlalchemy.engine.reflection import Inspector

from datahub.ingestion.api.closeable import Closeable
from datahub.ingestion.source.unity.proxy_types import (
    Catalog,
    Column,
    ColumnProfile,
    CustomCatalogType,
    HiveTableType,
    Metastore,
    Schema,
    Table,
    TableProfile,
    TableReference,
)
from datahub.ingestion.source.unity.report import UnityCatalogReport

logger = logging.getLogger(__name__)
HIVE_METASTORE = "hive_metastore"

type_map = {
    "boolean": ColumnTypeName.BOOLEAN,
    "tinyint": ColumnTypeName.INT,
    "smallint": ColumnTypeName.INT,
    "int": ColumnTypeName.INT,
    "bigint": ColumnTypeName.LONG,
    "float": ColumnTypeName.FLOAT,
    "double": ColumnTypeName.DOUBLE,
    "decimal": ColumnTypeName.DECIMAL,
    "string": ColumnTypeName.STRING,
    "varchar": ColumnTypeName.STRING,
    "timestamp": ColumnTypeName.TIMESTAMP,
    "date": ColumnTypeName.DATE,
    "binary": ColumnTypeName.BINARY,
}

NUM_NULLS = "num_nulls"
DISTINCT_COUNT = "distinct_count"
MIN = "min"
MAX = "max"
AVG_COL_LEN = "avg_col_len"
MAX_COL_LEN = "max_col_len"
VERSION = "version"

ROWS = "rows"
BYTES = "bytes"
TABLE_STAT_LIST = {ROWS, BYTES}


class HiveMetastoreProxy(Closeable):
    # Why not use hive ingestion source directly here ?
    # 1. hive ingestion source assumes 2-level namespace heirarchy and currently
    #    there is no other intermediate interface except sqlalchemy inspector
    #    that can be used to fetch hive metadata.
    # 2. hive recipe for databricks (databricks+pyhive dialect) does not
    #    readily support SQL warehouse. Also this dialect is not actively maintained.
    """
    Proxy to read metadata from hive_metastore databricks catalog. This is required
    as unity catalog apis do not return details about this legacy metastore.
    """

    def __init__(
        self, sqlalchemy_url: str, options: dict, report: UnityCatalogReport
    ) -> None:
        try:
            self.inspector = HiveMetastoreProxy.get_inspector(sqlalchemy_url, options)
            self.report = report
        except Exception:
            # This means that there is no `hive_metastore` catalog in databricks workspace
            # Not tested but seems like the logical conclusion.
            raise

    @staticmethod
    def get_inspector(sqlalchemy_url: str, options: dict) -> Inspector:
        engine = create_engine(sqlalchemy_url, **options)
        return inspect(engine.connect())

    def hive_metastore_catalog(self, metastore: Optional[Metastore]) -> Catalog:
        return Catalog(
            id=f"{metastore.id}.{HIVE_METASTORE}" if metastore else HIVE_METASTORE,
            name=HIVE_METASTORE,
            comment=None,
            metastore=metastore,
            owner=None,
            type=CustomCatalogType.HIVE_METASTORE_CATALOG,
        )

    def hive_metastore_schemas(self, catalog: Catalog) -> Iterable[Schema]:
        for schema_name in self.inspector.get_schema_names():
            yield Schema(
                name=schema_name,
                id=f"{catalog.id}.{schema_name}",
                catalog=catalog,
                comment=None,
                owner=None,
            )

    def hive_metastore_tables(self, schema: Schema) -> Iterable[Table]:
        # NOTE: Ideally, we use `inspector.get_view_names` and `inspector.get_table_names` here instead of
        # making show queries in this class however Databricks dialect for databricks-sql-connector<3.0.0 does not
        # back-quote schemas with special char such as hyphen.
        # Currently, databricks-sql-connector is pinned to <3.0.0 due to requirement of SQLAlchemy > 2.0.21 for
        # later versions.
        views = self.get_view_names(schema.name)
        for table_name in views:
            yield self._get_table(schema, table_name, True)

        for table_name in self.get_table_names(schema.name):
            if table_name in views:
                continue
            yield self._get_table(schema, table_name, False)

    def get_table_names(self, schema_name: str) -> List[str]:
        try:
            rows = self._execute_sql(f"SHOW TABLES FROM `{schema_name}`")
            # 3 columns - database, tableName, isTemporary
            return [row.tableName for row in rows]
        except Exception as e:
            self.report.report_warning(
                "Failed to get tables for schema", f"{HIVE_METASTORE}.{schema_name}"
            )
            logger.warning(
                f"Failed to get tables {schema_name} due to {e}", exc_info=True
            )
        return []

    def get_view_names(self, schema_name: str) -> List[str]:
        try:
            rows = self._execute_sql(f"SHOW VIEWS FROM `{schema_name}`")
            # 3 columns - database, tableName, isTemporary
            return [row.tableName for row in rows]
        except Exception as e:
            self.report.report_warning("Failed to get views for schema", schema_name)
            logger.warning(
                f"Failed to get views {schema_name} due to {e}", exc_info=True
            )
        return []

    def _get_table(
        self,
        schema: Schema,
        table_name: str,
        is_view: bool = False,
    ) -> Table:
        columns = self._get_columns(schema.name, table_name)
        detailed_info = self._get_table_info(schema.name, table_name)

        comment = detailed_info.pop("Comment", None)
        storage_location = detailed_info.pop("Location", None)
        datasource_format = self._get_datasource_format(
            detailed_info.pop("Provider", None)
        )

        created_at = self._get_created_at(detailed_info.pop("Created Time", None))

        return Table(
            name=table_name,
            id=f"{schema.id}.{table_name}",
            table_type=self._get_table_type(detailed_info.pop("Type", None)),
            schema=schema,
            columns=columns,
            storage_location=storage_location,
            data_source_format=datasource_format,
            view_definition=(
                self._get_view_definition(schema.name, table_name) if is_view else None
            ),
            properties=detailed_info,
            owner=None,
            generation=None,
            created_at=created_at,
            created_by=None,
            updated_at=None,
            updated_by=None,
            table_id=f"{schema.id}.{table_name}",
            comment=comment,
        )

    def get_table_profile(
        self, ref: TableReference, include_column_stats: bool = False
    ) -> Optional[TableProfile]:
        columns = self._get_columns(
            ref.schema,
            ref.table,
        )
        detailed_info = self._get_table_info(ref.schema, ref.table)

        if not columns and not detailed_info:
            return None

        table_stats = (
            self._get_cached_table_statistics(detailed_info["Statistics"])
            if detailed_info.get("Statistics")
            else {}
        )

        column_profiles: List[ColumnProfile] = []
        if include_column_stats:
            for column in columns:
                column_profile = self._get_column_profile(column.name, ref)
                if column_profile:
                    column_profiles.append(column_profile)

        return TableProfile(
            num_rows=(
                int(table_stats[ROWS]) if table_stats.get(ROWS) is not None else None
            ),
            total_size=(
                int(table_stats[BYTES]) if table_stats.get(BYTES) is not None else None
            ),
            num_columns=len(columns),
            column_profiles=column_profiles,
        )

    def _get_column_profile(
        self, column: str, ref: TableReference
    ) -> Optional[ColumnProfile]:
        try:
            props = self._column_describe_extended(ref.schema, ref.table, column)
            col_stats = {}
            for prop in props:
                col_stats[prop[0]] = prop[1]
            return ColumnProfile(
                name=column,
                null_count=(
                    int(col_stats[NUM_NULLS])
                    if col_stats.get(NUM_NULLS) is not None
                    else None
                ),
                distinct_count=(
                    int(col_stats[DISTINCT_COUNT])
                    if col_stats.get(DISTINCT_COUNT) is not None
                    else None
                ),
                min=col_stats.get(MIN),
                max=col_stats.get(MAX),
                avg_len=col_stats.get(AVG_COL_LEN),
                max_len=col_stats.get(MAX_COL_LEN),
                version=col_stats.get(VERSION),
            )
        except Exception as e:
            logger.debug(f"Failed to get column profile for {ref}.{column} due to {e}")
            return None

    def _get_cached_table_statistics(self, statistics: str) -> dict:
        # statistics is in format "xx bytes" OR "1382 bytes, 2 rows"
        table_stats = dict()
        for prop in statistics.split(","):
            value_key_list = prop.strip().split(" ")  # value_key_list -> [value, key]
            if len(value_key_list) == 2 and value_key_list[1] in TABLE_STAT_LIST:
                table_stats[value_key_list[1]] = value_key_list[0]

        return table_stats

    def _get_created_at(self, created_at: Optional[str]) -> Optional[datetime]:
        return (
            datetime.strptime(created_at, "%a %b %d %H:%M:%S %Z %Y")
            if created_at
            else None
        )

    def _get_datasource_format(
        self, provider: Optional[str]
    ) -> Optional[DataSourceFormat]:
        raw_format = provider
        if raw_format:
            try:
                return DataSourceFormat(raw_format.upper())
            except Exception:
                logger.debug(f"Unknown datasource format : {raw_format}")
                pass
        return None

    def _get_view_definition(self, schema_name: str, table_name: str) -> Optional[str]:
        try:
            rows = self._execute_sql(
                f"SHOW CREATE TABLE `{schema_name}`.`{table_name}`"
            )
            for row in rows:
                return row[0]
        except Exception as e:
            self.report.report_warning(
                "Failed to get view definition for table",
                f"{HIVE_METASTORE}.{schema_name}.{table_name}",
            )
            logger.debug(
                f"Failed to get view definition for {schema_name}.{table_name} due to {e}",
                exc_info=True,
            )
        return None

    def _get_table_type(self, type: Optional[str]) -> HiveTableType:
        if type == "EXTERNAL":
            return HiveTableType.HIVE_EXTERNAL_TABLE
        elif type == "MANAGED":
            return HiveTableType.HIVE_MANAGED_TABLE
        elif type == "VIEW":
            return HiveTableType.HIVE_VIEW
        else:
            return HiveTableType.UNKNOWN

    @lru_cache(maxsize=1)
    def _get_table_info(self, schema_name: str, table_name: str) -> dict:
        # Generate properties dictionary.
        properties = {}

        try:
            rows = self._describe_extended(schema_name, table_name)

            index = rows.index(("# Detailed Table Information", "", ""))
            rows = rows[index + 1 :]
            # Copied from https://github.com/acryldata/PyHive/blob/master/pyhive/sqlalchemy_hive.py#L375

            active_heading = None
            for col_name, data_type, value in rows:
                col_name = col_name.rstrip()
                if col_name.startswith("# "):
                    continue
                elif col_name == "" and data_type is None:
                    active_heading = None
                    continue
                elif col_name != "" and data_type is None:
                    active_heading = col_name
                elif col_name != "" and data_type is not None:
                    properties[col_name] = data_type.strip()
                else:
                    # col_name == "", data_type is not None
                    prop_name = f"{active_heading} {data_type.rstrip()}"
                    properties[prop_name] = value.rstrip()
        except Exception as e:
            self.report.report_warning(
                "Failed to get detailed info for table",
                f"{HIVE_METASTORE}.{schema_name}.{table_name}",
            )
            logger.debug(
                f"Failed to get detailed info for table {schema_name}.{table_name} due to {e}",
                exc_info=True,
            )
        return properties

    @lru_cache(maxsize=1)
    def _get_columns(self, schema_name: str, table_name: str) -> List[Column]:
        columns: List[Column] = []
        try:
            rows = self._describe_extended(schema_name, table_name)
            for i, row in enumerate(rows):
                if i == 0 and row[0].strip() == "col_name":
                    continue  # first row
                if row[0].strip() in (
                    "",
                    "# Partition Information",
                    "# Detailed Table Information",
                ):
                    break
                columns.append(
                    Column(
                        name=row[0].strip(),
                        id=f"{HIVE_METASTORE}.{schema_name}.{table_name}.{row[0].strip()}",
                        type_text=row[1].strip(),
                        type_name=type_map.get(row[1].strip().lower()),
                        type_scale=None,
                        type_precision=None,
                        position=None,
                        nullable=None,
                        comment=row[2],
                    )
                )
        except Exception as e:
            self.report.report_warning(
                "Failed to get columns for table",
                f"{HIVE_METASTORE}.{schema_name}.{table_name}",
            )
            logger.debug(
                f"Failed to get columns for table {schema_name}.{table_name} due to {e}",
                exc_info=True,
            )
        return columns

    @lru_cache(maxsize=1)
    def _describe_extended(self, schema_name: str, table_name: str) -> List[Row]:
        """
        Rows are structured as shown in examples here
        https://docs.databricks.com/en/sql/language-manual/sql-ref-syntax-aux-describe-table.html#examples
        """
        return self._execute_sql(f"DESCRIBE EXTENDED `{schema_name}`.`{table_name}`")

    def _column_describe_extended(
        self, schema_name: str, table_name: str, column_name: str
    ) -> List[Row]:
        """
        Rows are structured as shown in examples here
        https://docs.databricks.com/en/sql/language-manual/sql-ref-syntax-aux-describe-table.html#examples
        """
        return self._execute_sql(
            f"DESCRIBE EXTENDED `{schema_name}`.`{table_name}` {column_name}"
        )

    def _execute_sql(self, sql: str) -> List[Row]:
        return self.inspector.bind.execute(sql).fetchall()

    def close(self):
        self.inspector.bind.close()  # type:ignore



import logging
import sys
from typing import Any, Dict, List, Optional, Union

import requests

from datahub.ingestion.source.sigma.config import Constant, SigmaSourceConfig
from datahub.ingestion.source.sigma.data_classes import (
    Element,
    Page,
    SigmaDataset,
    Workbook,
    Workspace,
)

# Logger instance
logger = logging.getLogger(__name__)


class SigmaAPI:
    def __init__(self, config: SigmaSourceConfig) -> None:
        self.config = config
        self.workspaces: Dict[str, Workspace] = {}
        self.users: Dict[str, str] = {}
        self.session = requests.Session()
        # Test connection by generating access token
        logger.info(f"Trying to connect to {self.config.api_url}")
        self._generate_token()

    def _generate_token(self):
        data = {
            "grant_type": "client_credentials",
            "client_id": self.config.client_id,
            "client_secret": self.config.client_secret,
        }
        response = self.session.post(f"{self.config.api_url}/auth/token", data=data)
        response.raise_for_status()
        self.session.headers.update(
            {
                "Authorization": f"Bearer {response.json()['access_token']}",
                "Content-Type": "application/json",
            }
        )

    def _log_http_error(self, message: str) -> Any:
        logger.warning(message)
        _, e, _ = sys.exc_info()
        if isinstance(e, requests.exceptions.HTTPError):
            logger.warning(f"HTTP status-code = {e.response.status_code}")
        logger.debug(msg=message, exc_info=e)
        return e

    def get_workspace(self, workspace_id: str) -> Optional[Workspace]:
        workspace: Optional[Workspace] = None
        try:
            response = self.session.get(
                f"{self.config.api_url}/workspaces/{workspace_id}"
            )
            response.raise_for_status()
            workspace_dict = response.json()
            workspace = Workspace.parse_obj(workspace_dict)
        except Exception as e:
            self._log_http_error(
                message=f"Unable to fetch workspace {workspace_id}. Exception: {e}"
            )
        return workspace

    def get_user_name(self, user_id: str) -> Optional[str]:
        try:
            if user_id in self.users:
                # To avoid fetching same user details again
                return self.users[user_id]
            else:
                response = self.session.get(f"{self.config.api_url}/members/{user_id}")
                response.raise_for_status()
                user_dict = response.json()
                user_name = (
                    f"{user_dict[Constant.FIRSTNAME]}_{user_dict[Constant.LASTNAME]}"
                )
                self.users[user_id] = user_name
                return user_name
        except Exception as e:
            self._log_http_error(
                message=f"Unable to fetch user with id {user_id}. Exception: {e}"
            )
        return None

    def get_sigma_dataset(
        self, dataset_id: str, workspace_id: str, path: str
    ) -> Optional[SigmaDataset]:
        dataset: Optional[SigmaDataset] = None
        try:
            response = self.session.get(f"{self.config.api_url}/datasets/{dataset_id}")
            response.raise_for_status()
            dataset_dict = response.json()
            dataset_dict[Constant.WORKSPACEID] = workspace_id
            dataset_dict[Constant.PATH] = path
            dataset = SigmaDataset.parse_obj(dataset_dict)
        except Exception as e:
            self._log_http_error(
                message=f"Unable to fetch sigma dataset {dataset_id}. Exception: {e}"
            )
        return dataset

    def _get_element_upstream_sources(
        self, element_id: str, workbook_id: str
    ) -> Dict[str, str]:
        """
        Returns upstream dataset sources with keys as id and values as name of that dataset
        """
        upstream_sources: Dict[str, str] = {}
        try:
            response = self.session.get(
                f"{self.config.api_url}/workbooks/{workbook_id}/lineage/elements/{element_id}"
            )
            response.raise_for_status()
            response_dict = response.json()
            for edge in response_dict[Constant.EDGES]:
                source_type = response_dict[Constant.DEPENDENCIES][
                    edge[Constant.SOURCE]
                ][Constant.TYPE]
                if source_type == "dataset":
                    upstream_sources[edge[Constant.SOURCE]] = response_dict[
                        Constant.DEPENDENCIES
                    ][edge[Constant.SOURCE]][Constant.NAME]
        except Exception as e:
            self._log_http_error(
                message=f"Unable to fetch lineage of element {element_id}. Exception: {e}"
            )
        return upstream_sources

    def _get_element_sql_query(
        self, element_id: str, workbook_id: str
    ) -> Optional[str]:
        query: Optional[str] = None
        try:
            response = self.session.get(
                f"{self.config.api_url}/workbooks/{workbook_id}/elements/{element_id}/query"
            )
            response.raise_for_status()
            response_dict = response.json()
            if "sql" in response_dict:
                query = response_dict["sql"]
        except Exception as e:
            self._log_http_error(
                message=f"Unable to fetch sql query for a element {element_id}. Exception: {e}"
            )
        return query

    def get_page_elements(self, workbook: Workbook, page: Page) -> List[Element]:
        elements: List[Element] = []
        try:
            response = self.session.get(
                f"{self.config.api_url}/workbooks/{workbook.workbookId}/pages/{page.pageId}/elements"
            )
            response.raise_for_status()
            for i, element_dict in enumerate(response.json()[Constant.ENTRIES]):
                if not element_dict.get(Constant.NAME):
                    element_dict[Constant.NAME] = f"Element {i+1} of Page '{page.name}'"
                element_dict[
                    Constant.URL
                ] = f"{workbook.url}?:nodeId={element_dict[Constant.ELEMENTID]}&:fullScreen=true"
                element = Element.parse_obj(element_dict)
                element.upstream_sources = self._get_element_upstream_sources(
                    element.elementId, workbook.workbookId
                )
                element.query = self._get_element_sql_query(
                    element.elementId, workbook.workbookId
                )
                elements.append(element)
        except Exception as e:
            self._log_http_error(
                message=f"Unable to fetch elements of page {page.pageId}, workbook {workbook.workbookId}. Exception: {e}"
            )
        return elements

    def get_workbook_pages(self, workbook: Workbook) -> List[Page]:
        pages: List[Page] = []
        try:
            response = self.session.get(
                f"{self.config.api_url}/workbooks/{workbook.workbookId}/pages"
            )
            response.raise_for_status()
            for page_dict in response.json()[Constant.ENTRIES]:
                page = Page.parse_obj(page_dict)
                page.elements = self.get_page_elements(workbook, page)
                pages.append(page)
        except Exception as e:
            self._log_http_error(
                message=f"Unable to fetch pages of workbook {workbook.workbookId}. Exception: {e}"
            )
        return pages

    def get_workbook(self, workbook_id: str, workspace_id: str) -> Optional[Workbook]:
        workbook: Optional[Workbook] = None
        try:
            response = self.session.get(
                f"{self.config.api_url}/workbooks/{workbook_id}"
            )
            response.raise_for_status()
            workbook_dict = response.json()
            workbook_dict[Constant.WORKSPACEID] = workspace_id
            workbook = Workbook.parse_obj(workbook_dict)
            workbook.pages = self.get_workbook_pages(workbook)
        except Exception as e:
            self._log_http_error(
                message=f"Unable to fetch workbook {workbook_id}. Exception: {e}"
            )
        return workbook

    def get_workspace_id(self, parent_id: str, path: str) -> str:
        path_list = path.split("/")
        while len(path_list) != 1:  # means current parent id is folder's id
            response = self.session.get(f"{self.config.api_url}/files/{parent_id}")
            parent_id = response.json()[Constant.PARENTID]
            path_list.pop()
        return parent_id

    def get_sigma_entities(self) -> List[Union[Workbook, SigmaDataset]]:
        entities: List[Union[Workbook, SigmaDataset]] = []
        url = f"{self.config.api_url}/files"
        while True:
            response = self.session.get(url)
            response.raise_for_status()
            response_dict = response.json()
            for entity in response_dict[Constant.ENTRIES]:
                workspace_id = self.get_workspace_id(
                    entity[Constant.PARENTID], entity[Constant.PATH]
                )
                if workspace_id not in self.workspaces:
                    workspace = self.get_workspace(workspace_id)
                    if workspace:
                        self.workspaces[workspace.workspaceId] = workspace

                if self.workspaces.get(
                    workspace_id
                ) and self.config.workspace_pattern.allowed(
                    self.workspaces[workspace_id].name
                ):
                    type = entity[Constant.TYPE]
                    if type == Constant.DATASET:
                        dataset = self.get_sigma_dataset(
                            entity[Constant.ID],
                            workspace_id,
                            entity[Constant.PATH],
                        )
                        if dataset:
                            dataset.badge = entity[Constant.BADGE]
                            entities.append(dataset)
                    elif type == Constant.WORKBOOK:
                        workbook = self.get_workbook(entity[Constant.ID], workspace_id)
                        if workbook:
                            workbook.badge = entity[Constant.BADGE]
                            entities.append(workbook)
            if response_dict[Constant.NEXTPAGE]:
                url = f"{url}?page={response_dict[Constant.NEXTPAGE]}"
            else:
                break
        return entities



import logging
from typing import List, Optional, Tuple

from lark import Tree

from datahub.ingestion.source.powerbi.m_query import resolver, tree_function

logger = logging.getLogger(__name__)


def validate_parse_tree(
    tree: Tree, native_query_enabled: bool = True
) -> Tuple[bool, Optional[str]]:
    """
    :param tree: tree to validate as per functions supported by m_parser module
    :param native_query_enabled: Whether user want to extract lineage from native query
    :return: first argument is False if validation is failed and second argument would contain the error message.
             in-case of valid tree the first argument is True and second argument would be None.
    """
    functions: List[str] = tree_function.get_all_function_name(tree)
    if len(functions) == 0:
        return False, "Function calls not found"

    if native_query_enabled is False:
        if resolver.FunctionName.NATIVE_QUERY.value in functions:
            return False, "Lineage extraction from native query is disabled."

    return True, None



import dataclasses
import json
import logging
import pprint
from dataclasses import dataclass
from datetime import datetime, timedelta
from enum import Enum
from typing import Any, Optional, runtime_checkable

import humanfriendly
import pydantic
from pydantic import BaseModel
from typing_extensions import Literal, Protocol

from datahub.ingestion.api.report_helpers import format_datetime_relative
from datahub.utilities.lossy_collections import LossyList

logger = logging.getLogger(__name__)
LogLevel = Literal["ERROR", "WARNING", "INFO", "DEBUG"]


@runtime_checkable
class SupportsAsObj(Protocol):
    def as_obj(self) -> dict:
        ...


@dataclass
class Report(SupportsAsObj):
    @staticmethod
    def to_str(some_val: Any) -> str:
        if isinstance(some_val, Enum):
            return some_val.name
        else:
            return str(some_val)

    @staticmethod
    def to_pure_python_obj(some_val: Any) -> Any:
        """A cheap way to generate a dictionary."""

        if isinstance(some_val, SupportsAsObj):
            return some_val.as_obj()
        elif isinstance(some_val, pydantic.BaseModel):
            return Report.to_pure_python_obj(some_val.dict())
        elif dataclasses.is_dataclass(some_val):
            return dataclasses.asdict(some_val)
        elif isinstance(some_val, list):
            return [Report.to_pure_python_obj(v) for v in some_val if v is not None]
        elif isinstance(some_val, timedelta):
            return humanfriendly.format_timespan(some_val)
        elif isinstance(some_val, datetime):
            try:
                return format_datetime_relative(some_val)
            except Exception:
                # we don't want to fail reporting because we were unable to pretty print a timestamp
                return str(datetime)
        elif isinstance(some_val, dict):
            return {
                Report.to_str(k): Report.to_pure_python_obj(v)
                for k, v in some_val.items()
                if v is not None
            }
        elif isinstance(some_val, (int, float, bool)):
            return some_val
        else:
            # fall through option
            return Report.to_str(some_val)

    def compute_stats(self) -> None:
        """A hook to compute derived stats"""
        pass

    def as_obj(self) -> dict:
        self.compute_stats()
        return {
            str(key): Report.to_pure_python_obj(value)
            for (key, value) in self.__dict__.items()
            # ignore nulls and fields starting with _
            if value is not None and not str(key).startswith("_")
        }

    def as_string(self) -> str:
        return pprint.pformat(self.as_obj(), width=150, sort_dicts=False)

    def as_json(self) -> str:
        return json.dumps(self.as_obj())

    # TODO add helper method for warning / failure status + counts?


class ReportAttribute(BaseModel):
    severity: LogLevel = "DEBUG"
    help: Optional[str] = None

    @property
    def logger_sev(self) -> int:
        log_levels = {
            "DEBUG": logging.DEBUG,
            "INFO": logging.INFO,
            "WARNING": logging.WARNING,
            "ERROR": logging.ERROR,
        }
        return log_levels[self.severity]

    def log(self, msg: str) -> None:
        logger.log(level=self.logger_sev, msg=msg, stacklevel=3)


class EntityFilterReport(ReportAttribute):
    type: str

    processed_entities: LossyList[str] = pydantic.Field(default_factory=LossyList)
    dropped_entities: LossyList[str] = pydantic.Field(default_factory=LossyList)

    def processed(self, entity: str, type: Optional[str] = None) -> None:
        self.log(f"Processed {type or self.type} {entity}")
        self.processed_entities.append(entity)

    def dropped(self, entity: str, type: Optional[str] = None) -> None:
        self.log(f"Filtered {type or self.type} {entity}")
        self.dropped_entities.append(entity)

    def as_obj(self) -> dict:
        return {
            "filtered": self.dropped_entities.as_obj(),
            "processed": self.processed_entities.as_obj(),
        }

    @staticmethod
    def field(type: str, severity: LogLevel = "DEBUG") -> "EntityFilterReport":
        """A helper to create a dataclass field."""

        return dataclasses.field(
            default_factory=lambda: EntityFilterReport(type=type, severity=severity)
        )



from typing import Optional


def unquote_and_decode_unicode_escape_seq(
    string: str,
    leading_quote: str = '"',
    trailing_quote: Optional[str] = None,
) -> str:
    """
    If string starts and ends with a quote, unquote it and decode Unicode escape sequences
    """
    trailing_quote = trailing_quote if trailing_quote else leading_quote

    if string.startswith(leading_quote) and string.endswith(trailing_quote):
        string = string[1:-1]

    cleaned_string = string.encode().decode("unicode-escape")

    return cleaned_string



from functools import lru_cache
from typing import List, Optional, Tuple, TypedDict

from datahub.api.entities.assertion.assertion import BaseEntityAssertion
from datahub.ingestion.graph.client import get_default_graph
from datahub.metadata.com.linkedin.pegasus2avro.dataset import DatasetProperties
from datahub.metadata.com.linkedin.pegasus2avro.schema import SchemaMetadata
from datahub.utilities.urns.urn import Urn


class ColumnDict(TypedDict):
    col: str
    native_type: str


@lru_cache
def get_qualified_name_from_datahub(urn: str) -> Optional[str]:
    with get_default_graph() as graph:
        props: Optional[DatasetProperties] = graph.get_aspect(urn, DatasetProperties)
        if props is not None:
            return props.qualifiedName
    return None


@lru_cache
def get_schema_from_datahub(urn: str) -> Optional[List[ColumnDict]]:
    with get_default_graph() as graph:
        schema: Optional[SchemaMetadata] = graph.get_aspect(urn, SchemaMetadata)
        if schema is not None:
            return [
                {"col": field.fieldPath, "native_type": field.nativeDataType}
                for field in schema.fields
            ]
    return None


def get_entity_name(assertion: BaseEntityAssertion) -> Tuple[str, str, str]:
    if assertion.meta and assertion.meta.get("entity_qualified_name"):
        parts = assertion.meta["entity_qualified_name"].split(".")
    else:
        qualified_name = get_qualified_name_from_datahub(assertion.entity)
        if qualified_name is not None:
            parts = qualified_name.split(".")
        else:
            urn_id = Urn.create_from_string(assertion.entity).entity_ids[1]
            parts = urn_id.split(".")
            if len(parts) > 3:
                parts = parts[-3:]
    assert len(parts) == 3
    database = parts[-3]
    schema = parts[-2]
    table = parts[-1]
    return database, schema, table


def get_entity_schema(assertion: BaseEntityAssertion) -> Optional[List[ColumnDict]]:
    if assertion.meta and assertion.meta.get("entity_schema"):
        return assertion.meta.get("entity_schema")
    elif get_schema_from_datahub(assertion.entity):
        return get_schema_from_datahub(assertion.entity)
    return None



"""
For helper methods to contain manipulation of the config file in local system.
"""

import logging
import os
import sys
from typing import Optional, Union

import click
import yaml
from pydantic import BaseModel, ValidationError

from datahub.cli.env_utils import get_boolean_env_variable

log = logging.getLogger(__name__)

DEFAULT_GMS_HOST = "http://localhost:8080"
CONDENSED_DATAHUB_CONFIG_PATH = "~/.datahubenv"
DATAHUB_CONFIG_PATH = os.path.expanduser(CONDENSED_DATAHUB_CONFIG_PATH)
DATAHUB_ROOT_FOLDER = os.path.expanduser("~/.datahub")
ENV_SKIP_CONFIG = "DATAHUB_SKIP_CONFIG"


class GmsConfig(BaseModel):
    server: str
    token: Optional[str] = None


class DatahubConfig(BaseModel):
    gms: GmsConfig


def persist_datahub_config(config: dict) -> None:
    with open(DATAHUB_CONFIG_PATH, "w+") as outfile:
        yaml.dump(config, outfile, default_flow_style=False)
    return None


def write_gms_config(
    host: str, token: Optional[str], merge_with_previous: bool = True
) -> None:
    config = DatahubConfig(gms=GmsConfig(server=host, token=token))
    if merge_with_previous:
        try:
            previous_config = get_client_config(as_dict=True)
            assert isinstance(previous_config, dict)
        except Exception as e:
            # ok to fail on this
            previous_config = {}
            log.debug(
                f"Failed to retrieve config from file {DATAHUB_CONFIG_PATH}: {e}. This isn't fatal."
            )
        config_dict = {**previous_config, **config.dict()}
    else:
        config_dict = config.dict()
    persist_datahub_config(config_dict)


def get_details_from_config():
    datahub_config = get_client_config(as_dict=False)
    assert isinstance(datahub_config, DatahubConfig)
    if datahub_config is not None:
        gms_config = datahub_config.gms

        gms_host = gms_config.server
        gms_token = gms_config.token
        return gms_host, gms_token
    else:
        return None, None


def should_skip_config() -> bool:
    return get_boolean_env_variable(ENV_SKIP_CONFIG, False)


def ensure_datahub_config() -> None:
    if not os.path.isfile(DATAHUB_CONFIG_PATH):
        click.secho(
            f"No {CONDENSED_DATAHUB_CONFIG_PATH} file found, generating one for you...",
            bold=True,
        )
        write_gms_config(DEFAULT_GMS_HOST, None)


def get_client_config(as_dict: bool = False) -> Union[Optional[DatahubConfig], dict]:
    with open(DATAHUB_CONFIG_PATH) as stream:
        try:
            config_json = yaml.safe_load(stream)
            if as_dict:
                return config_json
            try:
                datahub_config = DatahubConfig.parse_obj(config_json)
                return datahub_config
            except ValidationError as e:
                click.echo(
                    f"Received error, please check your {CONDENSED_DATAHUB_CONFIG_PATH}"
                )
                click.echo(e, err=True)
                sys.exit(1)
        except yaml.YAMLError as exc:
            click.secho(f"{DATAHUB_CONFIG_PATH} malformed, error: {exc}", bold=True)
            return None



import logging
from abc import ABCMeta
from typing import List, Optional, cast

from datahub.configuration.common import (
    TransformerSemantics,
    TransformerSemanticsConfigModel,
)
from datahub.emitter.mce_builder import Aspect
from datahub.ingestion.graph.client import DataHubGraph
from datahub.ingestion.transformer.base_transformer import (
    BaseTransformer,
    SingleAspectTransformer,
)
from datahub.metadata.schema_classes import GlobalTagsClass

log = logging.getLogger(__name__)


class DatasetTransformer(BaseTransformer, SingleAspectTransformer, metaclass=ABCMeta):
    """Transformer that does transform sequentially on each dataset."""

    def __init__(self):
        super().__init__()

    def entity_types(self) -> List[str]:
        return ["dataset"]


class DatasetOwnershipTransformer(DatasetTransformer, metaclass=ABCMeta):
    def aspect_name(self) -> str:
        return "ownership"


class DatasetDomainTransformer(DatasetTransformer, metaclass=ABCMeta):
    def aspect_name(self) -> str:
        return "domains"


class DatasetStatusTransformer(DatasetTransformer, metaclass=ABCMeta):
    def aspect_name(self) -> str:
        return "status"


class DatasetTagsTransformer(DatasetTransformer, metaclass=ABCMeta):
    def aspect_name(self) -> str:
        return "globalTags"

    @staticmethod
    def merge_with_server_global_tags(
        graph: DataHubGraph, urn: str, global_tags_aspect: Optional[GlobalTagsClass]
    ) -> Optional[GlobalTagsClass]:
        if not global_tags_aspect or not global_tags_aspect.tags:
            # nothing to add, no need to consult server
            return None

        # Merge the transformed tags with existing server tags.
        # The transformed tags takes precedence, which may change the tag context.
        server_global_tags_aspect = graph.get_tags(entity_urn=urn)
        if server_global_tags_aspect:
            global_tags_aspect.tags = list(
                {
                    **{tag.tag: tag for tag in server_global_tags_aspect.tags},
                    **{tag.tag: tag for tag in global_tags_aspect.tags},
                }.values()
            )

        return global_tags_aspect

    @staticmethod
    def update_if_keep_existing(
        config: TransformerSemanticsConfigModel,
        in_global_tags_aspect: GlobalTagsClass,
        out_global_tags_aspect: GlobalTagsClass,
    ) -> None:
        """Check if user want to keep existing tags"""
        if in_global_tags_aspect is not None and config.replace_existing is False:
            tags_seen = set()
            for item in in_global_tags_aspect.tags:
                if item.tag not in tags_seen:
                    out_global_tags_aspect.tags.append(item)
                    tags_seen.add(item.tag)

    @staticmethod
    def get_result_semantics(
        config: TransformerSemanticsConfigModel,
        graph: Optional[DataHubGraph],
        urn: str,
        out_global_tags_aspect: Optional[GlobalTagsClass],
    ) -> Optional[Aspect]:
        if config.semantics == TransformerSemantics.PATCH:
            assert graph
            return cast(
                Optional[Aspect],
                DatasetTagsTransformer.merge_with_server_global_tags(
                    graph, urn, out_global_tags_aspect
                ),
            )

        return cast(Aspect, out_global_tags_aspect)


class DatasetTermsTransformer(DatasetTransformer, metaclass=ABCMeta):
    def aspect_name(self) -> str:
        return "glossaryTerms"


class DatasetPropertiesTransformer(DatasetTransformer, metaclass=ABCMeta):
    def aspect_name(self) -> str:
        return "datasetProperties"


class DatasetBrowsePathsTransformer(DatasetTransformer, metaclass=ABCMeta):
    def aspect_name(self) -> str:
        return "browsePaths"


class DatasetSchemaMetadataTransformer(DatasetTransformer, metaclass=ABCMeta):
    def aspect_name(self) -> str:
        return "schemaMetadata"


class DatasetDataproductTransformer(DatasetTransformer, metaclass=ABCMeta):
    def aspect_name(self) -> str:
        return "dataProductProperties"


class DatasetUsageStatisticsTransformer(DatasetTransformer, metaclass=ABCMeta):
    def aspect_name(self) -> str:
        return "datasetUsageStatistics"



BQ_DATETIME_FORMAT = "%Y-%m-%dT%H:%M:%SZ"
BQ_DATE_SHARD_FORMAT = "%Y%m%d"

BQ_EXTERNAL_TABLE_URL_TEMPLATE = "https://console.cloud.google.com/bigquery?project={project}&ws=!1m5!1m4!4m3!1s{project}!2s{dataset}!3s{table}"
BQ_EXTERNAL_DATASET_URL_TEMPLATE = "https://console.cloud.google.com/bigquery?project={project}&ws=!1m4!1m3!3m2!1s{project}!2s{dataset}"



import functools
import json
import logging
import uuid
from textwrap import dedent
from typing import Any, Dict, Iterable, List, Optional, Union

import sqlalchemy
import trino
from packaging import version
from pydantic.fields import Field
from sqlalchemy import exc, sql
from sqlalchemy.engine import reflection
from sqlalchemy.engine.base import Engine
from sqlalchemy.engine.reflection import Inspector
from sqlalchemy.sql import sqltypes
from sqlalchemy.types import TypeEngine
from trino.sqlalchemy import datatype
from trino.sqlalchemy.dialect import TrinoDialect

from datahub.configuration.source_common import (
    EnvConfigMixin,
    PlatformInstanceConfigMixin,
)
from datahub.emitter.mce_builder import make_dataset_urn_with_platform_instance
from datahub.emitter.mcp import MetadataChangeProposalWrapper
from datahub.ingestion.api.common import PipelineContext
from datahub.ingestion.api.decorators import (
    SourceCapability,
    SupportStatus,
    capability,
    config_class,
    platform_name,
    support_status,
)
from datahub.ingestion.api.workunit import MetadataWorkUnit
from datahub.ingestion.extractor import schema_util
from datahub.ingestion.source.common.data_reader import DataReader
from datahub.ingestion.source.sql.sql_common import (
    SQLAlchemySource,
    SqlWorkUnit,
    register_custom_type,
)
from datahub.ingestion.source.sql.sql_config import (
    BasicSQLAlchemyConfig,
    SQLCommonConfig,
)
from datahub.metadata.com.linkedin.pegasus2avro.common import Siblings
from datahub.metadata.com.linkedin.pegasus2avro.dataset import (
    DatasetLineageType,
    Upstream,
    UpstreamLineage,
)
from datahub.metadata.com.linkedin.pegasus2avro.schema import (
    MapTypeClass,
    NumberTypeClass,
    RecordTypeClass,
    SchemaField,
)

register_custom_type(datatype.ROW, RecordTypeClass)
register_custom_type(datatype.MAP, MapTypeClass)
register_custom_type(datatype.DOUBLE, NumberTypeClass)


KNOWN_CONNECTOR_PLATFORM_MAPPING = {
    "clickhouse": "clickhouse",
    "hive": "hive",
    "glue": "glue",
    "iceberg": "iceberg",
    "mysql": "mysql",
    "postgresql": "postgres",
    "redshift": "redshift",
    "bigquery": "bigquery",
    "snowflake_distributed": "snowflake",
    "snowflake_parallel": "snowflake",
    "snowflake_jdbc": "snowflake",
}

TWO_TIER_CONNECTORS = ["clickhouse", "hive", "glue", "mysql", "iceberg"]

PROPERTIES_TABLE_SUPPORTED_CONNECTORS = ["hive", "iceberg"]

# Type JSON was introduced in trino sqlalchemy dialect in version 0.317.0
if version.parse(trino.__version__) >= version.parse("0.317.0"):
    register_custom_type(datatype.JSON, RecordTypeClass)


@functools.lru_cache
def gen_catalog_connector_dict(engine: Engine) -> Dict[str, str]:
    query = dedent(
        """
        SELECT *
        FROM "system"."metadata"."catalogs"
    """
    ).strip()
    res = engine.execute(sql.text(query))
    return {row.catalog_name: row.connector_name for row in res}


def get_catalog_connector_name(engine: Engine, catalog_name: str) -> Optional[str]:
    return gen_catalog_connector_dict(engine).get(catalog_name)


# Read only table names and skip view names, as view names will also be returned
# from get_view_names
@reflection.cache  # type: ignore
def get_table_names(self, connection, schema: str = None, **kw):  # type: ignore
    schema = schema or self._get_default_schema_name(connection)
    if schema is None:
        raise exc.NoSuchTableError("schema is required")
    query = dedent(
        """
        SELECT "table_name"
        FROM "information_schema"."tables"
        WHERE "table_schema" = :schema and "table_type" != 'VIEW'
    """
    ).strip()
    res = connection.execute(sql.text(query), schema=schema)
    return [row.table_name for row in res]


# Include all table properties instead of only "comment" property
@reflection.cache  # type: ignore
def get_table_comment(self, connection, table_name: str, schema: str = None, **kw):  # type: ignore
    try:
        catalog_name = self._get_default_catalog_name(connection)
        if catalog_name is None:
            raise exc.NoSuchTableError("catalog is required in connection")
        connector_name = get_catalog_connector_name(connection.engine, catalog_name)
        if connector_name is None:
            return {}
        if connector_name in PROPERTIES_TABLE_SUPPORTED_CONNECTORS:
            properties_table = self._get_full_table(f"{table_name}$properties", schema)
            query = f"SELECT * FROM {properties_table}"
            row = connection.execute(sql.text(query)).fetchone()

            # Generate properties dictionary.
            properties = {}
            if row:
                for col_name, col_value in row.items():
                    if col_value is not None:
                        properties[col_name] = col_value

            return {"text": properties.get("comment", None), "properties": properties}
        else:
            return self.get_table_comment_default(connection, table_name, schema)
    except Exception:
        return {}


# Include column comment, original trino datatype as full_type
@reflection.cache  # type: ignore
def _get_columns(self, connection, table_name, schema: str = None, **kw):  # type: ignore
    schema = schema or self._get_default_schema_name(connection)
    query = dedent(
        """
        SELECT
            "column_name",
            "data_type",
            "column_default",
            UPPER("is_nullable") AS "is_nullable",
            "comment"
        FROM "information_schema"."columns"
        WHERE "table_schema" = :schema
            AND "table_name" = :table
        ORDER BY "ordinal_position" ASC
    """
    ).strip()
    res = connection.execute(sql.text(query), schema=schema, table=table_name)
    columns = []
    for record in res:
        column = dict(
            name=record.column_name,
            type=datatype.parse_sqltype(record.data_type),
            nullable=record.is_nullable == "YES",
            default=record.column_default,
            comment=record.comment,
        )
        columns.append(column)
    return columns


TrinoDialect.get_table_comment_default = TrinoDialect.get_table_comment
TrinoDialect.get_table_names = get_table_names
TrinoDialect.get_table_comment = get_table_comment
TrinoDialect._get_columns = _get_columns


class ConnectorDetail(PlatformInstanceConfigMixin, EnvConfigMixin):
    connector_database: Optional[str] = Field(default=None, description="")
    connector_platform: Optional[str] = Field(
        default=None,
        description="A connector's actual platform name. If not provided, will take from metadata tables"
        "Eg: hive catalog can have a connector platform as 'hive' or 'glue' or some other metastore.",
    )


class TrinoConfig(BasicSQLAlchemyConfig):
    # defaults
    scheme: str = Field(default="trino", description="", hidden_from_docs=True)
    database: str = Field(description="database (catalog)")

    catalog_to_connector_details: Dict[str, ConnectorDetail] = Field(
        default={},
        description="A mapping of trino catalog to its connector details like connector database, env and platform instance."
        "This configuration is used to build lineage to the underlying connector. Use catalog name as key.",
    )

    ingest_lineage_to_connectors: bool = Field(
        default=True,
        description="Whether lineage of datasets to connectors should be ingested",
    )

    trino_as_primary: bool = Field(
        default=True,
        description="Experimental feature. Whether trino dataset should be primary entity of the set of siblings",
    )

    def get_identifier(self: BasicSQLAlchemyConfig, schema: str, table: str) -> str:
        return f"{self.database}.{schema}.{table}"


@platform_name("Trino", doc_order=1)
@config_class(TrinoConfig)
@support_status(SupportStatus.CERTIFIED)
@capability(SourceCapability.DOMAINS, "Supported via the `domain` config field")
@capability(SourceCapability.DATA_PROFILING, "Optionally enabled via configuration")
class TrinoSource(SQLAlchemySource):
    """

    This plugin extracts the following:

    - Metadata for databases, schemas, and tables
    - Column types and schema associated with each table
    - Table, row, and column statistics via optional SQL profiling

    """

    config: TrinoConfig

    def __init__(
        self, config: TrinoConfig, ctx: PipelineContext, platform: str = "trino"
    ):
        super().__init__(config, ctx, platform)

    def get_db_name(self, inspector: Inspector) -> str:
        if self.config.database:
            return f"{self.config.database}"
        else:
            return super().get_db_name(inspector)

    def _get_source_dataset_urn(
        self,
        dataset_name: str,
        inspector: Inspector,
        schema: str,
        table: str,
    ) -> Optional[str]:
        catalog_name = dataset_name.split(".")[0]
        connector_name = get_catalog_connector_name(inspector.engine, catalog_name)
        if not connector_name:
            return None
        connector_details = self.config.catalog_to_connector_details.get(
            catalog_name, ConnectorDetail()
        )
        connector_platform_name = KNOWN_CONNECTOR_PLATFORM_MAPPING.get(
            connector_details.connector_platform or connector_name
        )
        if not connector_platform_name:
            logging.debug(f"Platform '{connector_platform_name}' is not yet supported.")
            return None

        if connector_platform_name in TWO_TIER_CONNECTORS:  # connector is two tier
            return make_dataset_urn_with_platform_instance(
                platform=connector_platform_name,
                name=f"{schema}.{table}",
                platform_instance=connector_details.platform_instance,
                env=connector_details.env,
            )
        elif connector_details.connector_database:  # else connector is three tier
            return make_dataset_urn_with_platform_instance(
                platform=connector_platform_name,
                name=f"{connector_details.connector_database}.{schema}.{table}",
                platform_instance=connector_details.platform_instance,
                env=connector_details.env,
            )
        else:
            logging.warning(f"Connector database missing for Catalog '{catalog_name}'.")
        return None

    def gen_siblings_workunit(
        self,
        dataset_urn: str,
        source_dataset_urn: str,
    ) -> Iterable[MetadataWorkUnit]:
        """
        Generate sibling workunit for both trino dataset and its connector source dataset
        """
        yield MetadataChangeProposalWrapper(
            entityUrn=dataset_urn,
            aspect=Siblings(
                primary=self.config.trino_as_primary, siblings=[source_dataset_urn]
            ),
        ).as_workunit()

        yield MetadataChangeProposalWrapper(
            entityUrn=source_dataset_urn,
            aspect=Siblings(
                primary=not self.config.trino_as_primary, siblings=[dataset_urn]
            ),
        ).as_workunit()

    def gen_lineage_workunit(
        self,
        dataset_urn: str,
        source_dataset_urn: str,
    ) -> Iterable[MetadataWorkUnit]:
        """
        Generate dataset to source connector lineage workunit
        """
        yield MetadataChangeProposalWrapper(
            entityUrn=dataset_urn,
            aspect=UpstreamLineage(
                upstreams=[
                    Upstream(dataset=source_dataset_urn, type=DatasetLineageType.VIEW)
                ]
            ),
        ).as_workunit()

    def _process_table(
        self,
        dataset_name: str,
        inspector: Inspector,
        schema: str,
        table: str,
        sql_config: SQLCommonConfig,
        data_reader: Optional[DataReader],
    ) -> Iterable[Union[SqlWorkUnit, MetadataWorkUnit]]:
        yield from super()._process_table(
            dataset_name, inspector, schema, table, sql_config, data_reader
        )
        if self.config.ingest_lineage_to_connectors:
            dataset_urn = make_dataset_urn_with_platform_instance(
                self.platform,
                dataset_name,
                self.config.platform_instance,
                self.config.env,
            )
            source_dataset_urn = self._get_source_dataset_urn(
                dataset_name, inspector, schema, table
            )
            if source_dataset_urn:
                yield from self.gen_siblings_workunit(dataset_urn, source_dataset_urn)
                yield from self.gen_lineage_workunit(dataset_urn, source_dataset_urn)

    def _process_view(
        self,
        dataset_name: str,
        inspector: Inspector,
        schema: str,
        view: str,
        sql_config: SQLCommonConfig,
    ) -> Iterable[Union[SqlWorkUnit, MetadataWorkUnit]]:
        yield from super()._process_view(
            dataset_name, inspector, schema, view, sql_config
        )
        if self.config.ingest_lineage_to_connectors:
            dataset_urn = make_dataset_urn_with_platform_instance(
                self.platform,
                dataset_name,
                self.config.platform_instance,
                self.config.env,
            )
            source_dataset_urn = self._get_source_dataset_urn(
                dataset_name, inspector, schema, view
            )
            if source_dataset_urn:
                yield from self.gen_siblings_workunit(dataset_urn, source_dataset_urn)

    @classmethod
    def create(cls, config_dict, ctx):
        config = TrinoConfig.parse_obj(config_dict)
        return cls(config, ctx)

    def get_schema_fields_for_column(
        self,
        dataset_name: str,
        column: dict,
        pk_constraints: Optional[dict] = None,
        tags: Optional[List[str]] = None,
    ) -> List[SchemaField]:
        fields = super().get_schema_fields_for_column(
            dataset_name, column, pk_constraints
        )

        if isinstance(column["type"], (datatype.ROW, sqltypes.ARRAY, datatype.MAP)):
            assert len(fields) == 1
            field = fields[0]
            # Get avro schema for subfields along with parent complex field
            avro_schema = self.get_avro_schema_from_data_type(
                column["type"], column["name"]
            )

            newfields = schema_util.avro_schema_to_mce_fields(
                json.dumps(avro_schema), default_nullable=True
            )

            # First field is the parent complex field
            newfields[0].nullable = field.nullable
            newfields[0].description = field.description
            newfields[0].isPartOfKey = field.isPartOfKey
            return newfields

        return fields

    def get_avro_schema_from_data_type(
        self, column_type: TypeEngine, column_name: str
    ) -> Dict[str, Any]:
        # Below Record structure represents the dataset level
        # Inner fields represent the complex field (struct/array/map/union)
        return {
            "type": "record",
            "name": "__struct_",
            "fields": [{"name": column_name, "type": _parse_datatype(column_type)}],
        }


_all_atomic_types = {
    sqltypes.BOOLEAN: "boolean",
    sqltypes.SMALLINT: "int",
    sqltypes.INTEGER: "int",
    sqltypes.BIGINT: "long",
    sqltypes.REAL: "float",
    datatype.DOUBLE: "double",  # type: ignore
    sqltypes.VARCHAR: "string",
    sqltypes.CHAR: "string",
    sqltypes.JSON: "record",
}


def _parse_datatype(s):
    if isinstance(s, sqlalchemy.types.ARRAY):
        return {
            "type": "array",
            "items": _parse_datatype(s.item_type),
            "native_data_type": repr(s),
        }
    elif isinstance(s, datatype.MAP):
        kt = _parse_datatype(s.key_type)
        vt = _parse_datatype(s.value_type)
        # keys are assumed to be strings in avro map
        return {
            "type": "map",
            "values": vt,
            "native_data_type": repr(s),
            "key_type": kt,
            "key_native_data_type": repr(s.key_type),
        }
    elif isinstance(s, datatype.ROW):
        return _parse_struct_fields(s.attr_types)
    else:
        return _parse_basic_datatype(s)


def _parse_struct_fields(parts):
    fields = []
    for name_and_type in parts:
        field_name = name_and_type[0].strip()
        field_type = _parse_datatype(name_and_type[1])
        fields.append({"name": field_name, "type": field_type})
    return {
        "type": "record",
        "name": "__struct_{}".format(str(uuid.uuid4()).replace("-", "")),
        "fields": fields,
        "native_data_type": f"ROW({parts})",
    }


def _parse_basic_datatype(s):
    for sql_type in _all_atomic_types.keys():
        if isinstance(s, sql_type):
            return {
                "type": _all_atomic_types[sql_type],
                "native_data_type": repr(s),
                "_nullable": True,
            }

    if isinstance(s, sqlalchemy.types.DECIMAL):
        return {
            "type": "bytes",
            "logicalType": "decimal",
            "precision": s.precision,  # type: ignore
            "scale": s.scale,  # type: ignore
            "native_data_type": repr(s),
            "_nullable": True,
        }
    elif isinstance(s, sqlalchemy.types.Date):
        return {
            "type": "int",
            "logicalType": "date",
            "native_data_type": repr(s),
            "_nullable": True,
        }
    elif isinstance(s, (sqlalchemy.types.DATETIME, sqlalchemy.types.TIMESTAMP)):
        return {
            "type": "int",
            "logicalType": "timestamp-millis",
            "native_data_type": repr(s),
            "_nullable": True,
        }

    return {"type": "null", "native_data_type": repr(s)}



import contextlib
import logging
import operator
import re
import time
from functools import reduce
from typing import Any, Dict, List, Mapping, Match, Optional, Union, cast

from datahub.configuration.common import ConfigModel
from datahub.emitter import mce_builder
from datahub.emitter.mce_builder import (
    OwnerType,
    make_user_urn,
    validate_ownership_type,
)
from datahub.metadata.schema_classes import (
    AuditStampClass,
    InstitutionalMemoryClass,
    InstitutionalMemoryMetadataClass,
    OwnerClass,
    OwnershipClass,
    OwnershipSourceClass,
    OwnershipTypeClass,
)

logger = logging.getLogger(__name__)


def _get_best_match(the_match: Match, group_name: str) -> str:
    with contextlib.suppress(IndexError):
        return the_match.group(group_name)

    with contextlib.suppress(IndexError):
        return the_match.group(1)

    return the_match.group(0)


def _make_owner_category_list(
    owner_type: OwnerType,
    owner_category: Any,
    owner_category_urn: Optional[str],
    owner_ids: List[str],
) -> List[Dict]:

    return [
        {
            "urn": mce_builder.make_owner_urn(owner_id, owner_type),
            "category": owner_category,
            "categoryUrn": owner_category_urn,
        }
        for owner_id in owner_ids
    ]


_match_regexp = re.compile(r"{{\s*\$match\s*}}", flags=re.MULTILINE)


def _insert_match_value(original_value: str, match_value: str) -> str:
    """
    If the original value is something like "foo{{ $match }}bar", then we insert the match value
    e.g. "foo<match_value>bar". Otherwise, it will leave the original value unchanged.
    """
    return _match_regexp.sub(match_value, original_value)


class Constants:
    ADD_DOC_LINK_OPERATION = "add_doc_link"
    ADD_TAG_OPERATION = "add_tag"
    ADD_TERM_OPERATION = "add_term"
    ADD_TERMS_OPERATION = "add_terms"
    ADD_OWNER_OPERATION = "add_owner"
    OPERATION = "operation"
    OPERATION_CONFIG = "config"
    TAG = "tag"
    TERM = "term"
    DOC_LINK = "link"
    DOC_DESCRIPTION = "description"
    OWNER_TYPE = "owner_type"
    OWNER_CATEGORY = "owner_category"
    MATCH = "match"
    USER_OWNER = "user"
    GROUP_OWNER = "group"
    OPERAND_DATATYPE_SUPPORTED = [int, bool, str, float]
    TAG_PARTITION_KEY = "PARTITION_KEY"
    TAG_DIST_KEY = "DIST_KEY"
    TAG_SORT_KEY = "SORT_KEY"
    SEPARATOR = "separator"


class _MappingOwner(ConfigModel):
    owner: str
    owner_type: str = OwnershipTypeClass.DATAOWNER


class _DatahubProps(ConfigModel):
    owners: List[Union[str, _MappingOwner]]

    def make_owner_category_list(self) -> List[Dict]:
        res = []
        for owner in self.owners:
            if isinstance(owner, str):
                owner_id = owner
                owner_category = OwnershipTypeClass.DATAOWNER
            else:
                owner_id = owner.owner
                owner_category = owner.owner_type
            owner_id = make_user_urn(owner_id)
            owner_category, owner_category_urn = validate_ownership_type(owner_category)

            res.append(
                {
                    "urn": owner_id,
                    "category": owner_category,
                    "categoryUrn": owner_category_urn,
                }
            )
        return res


class OperationProcessor:
    """
    A general class that processes a dictionary of properties and operations defined on it.
    An action is defined over a property with a specific match condition. If the condition is satisfied then the
    operation is carried out.
    Supported operations in this context include creating tags, terms or owners.
    For e.g below is a sample definition of two operations in yaml format
    operation_def:
      business_owner:
        match: ".*"
        operation: "add_owner"
        config:
          owner_type: "user"
      has_pii:
        match: true
        operation: "add_tag"
        config:
          tag: "has_pii"

    The raw input properties can be -
    "meta": {
        "business_owner": "jdoe.lastnew223@gmail.com",
        "has_pii": true,
      }
    If the match clause of both operations are satisfied on the raw properties a tag and a term aspect
    will be returned for further processing.
    """

    operation_defs: Dict[str, Dict] = {}
    tag_prefix: str = ""

    def __init__(
        self,
        operation_defs: Dict[str, Dict],
        tag_prefix: str = "",
        owner_source_type: Optional[str] = None,
        strip_owner_email_id: bool = False,
        match_nested_props: bool = False,
    ):
        self.operation_defs = operation_defs
        self.tag_prefix = tag_prefix
        self.strip_owner_email_id = strip_owner_email_id
        self.owner_source_type = owner_source_type
        self.match_nested_props = match_nested_props

    def process(self, raw_props: Mapping[str, Any]) -> Dict[str, Any]:  # noqa: C901
        # Defining the following local variables -
        # operations_map - the final resulting map when operations are processed.
        # Against each operation the values to be applied are stored.
        # for e.g "tag_operation" -> set("has_pii", "external")
        # operation_key : the property on which an operation is defined in the defs
        # operation config: map which contains the parameters to carry out that operation.
        # For e.g for add_tag operation config will have the tag value.
        # operation_type: the type of operation (add_tag, add_term, etc.)

        # Process the special "datahub" property, which supports tags, terms, and owners.
        operations_map: Dict[str, list] = {}
        try:
            datahub_prop = raw_props.get("datahub")
            if datahub_prop and isinstance(datahub_prop, dict):
                if datahub_prop.get("tags"):
                    # Note that tags get converted to urns later because we need to support the tag prefix.
                    tags = datahub_prop["tags"]
                    operations_map.setdefault(Constants.ADD_TAG_OPERATION, []).extend(
                        tags
                    )

                if datahub_prop.get("terms"):
                    terms = datahub_prop["terms"]
                    operations_map.setdefault(Constants.ADD_TERM_OPERATION, []).extend(
                        mce_builder.make_term_urn(term) for term in terms
                    )

                if datahub_prop.get("owners"):
                    owners = _DatahubProps.parse_obj_allow_extras(datahub_prop)
                    operations_map.setdefault(Constants.ADD_OWNER_OPERATION, []).extend(
                        owners.make_owner_category_list()
                    )
        except Exception as e:
            logger.error(f"Error while processing datahub property: {e}")

        # Process the actual directives.
        try:
            for operation_key in self.operation_defs:
                operation_type = self.operation_defs.get(operation_key, {}).get(
                    Constants.OPERATION
                )
                operation_config = self.operation_defs.get(operation_key, {}).get(
                    Constants.OPERATION_CONFIG
                )
                if not operation_type or not operation_config:
                    continue
                raw_props_value = raw_props.get(operation_key)
                if not raw_props_value and self.match_nested_props:
                    try:
                        raw_props_value = reduce(
                            operator.getitem, operation_key.split("."), raw_props
                        )
                    except KeyError:
                        pass

                maybe_match = self.get_match(
                    self.operation_defs[operation_key][Constants.MATCH],
                    raw_props_value,
                )
                if maybe_match is not None:
                    operation = self.get_operation_value(
                        operation_key, operation_type, operation_config, maybe_match
                    )

                    if operation_type == Constants.ADD_TERMS_OPERATION:
                        # add_terms operation is a special case where the operation value is a list of terms.
                        # We want to aggregate these values with the add_term operation.
                        operation_type = Constants.ADD_TERM_OPERATION

                    if operation:
                        if (
                            isinstance(operation, list)
                            and operation_type == Constants.ADD_OWNER_OPERATION
                        ):
                            operations_map.setdefault(operation_type, []).extend(
                                operation
                            )

                        elif isinstance(operation, (str, list)):
                            operations_map.setdefault(operation_type, []).extend(
                                operation
                                if isinstance(operation, list)
                                else [operation]
                            )
                        else:
                            operations_map.setdefault(operation_type, []).append(
                                operation
                            )
        except Exception as e:
            logger.error(f"Error while processing operation defs over raw_props: {e}")

        aspect_map: Dict[str, Any] = {}  # map of aspect name to aspect object
        try:
            aspect_map = self.convert_to_aspects(operations_map)
        except Exception as e:
            logger.error(f"Error while converting operations map to aspects: {e}")
        return aspect_map

    def convert_to_aspects(self, operation_map: Dict[str, list]) -> Dict[str, Any]:
        aspect_map: Dict[str, Any] = {}

        if Constants.ADD_TAG_OPERATION in operation_map:
            tag_aspect = mce_builder.make_global_tag_aspect_with_tag_list(
                sorted(set(operation_map[Constants.ADD_TAG_OPERATION]))
            )

            aspect_map[Constants.ADD_TAG_OPERATION] = tag_aspect

        if Constants.ADD_OWNER_OPERATION in operation_map:

            owner_aspect = OwnershipClass(
                owners=[
                    OwnerClass(
                        owner=x.get("urn"),
                        type=x.get("category"),
                        typeUrn=x.get("categoryUrn"),
                        source=OwnershipSourceClass(type=self.owner_source_type)
                        if self.owner_source_type
                        else None,
                    )
                    for x in sorted(
                        operation_map[Constants.ADD_OWNER_OPERATION],
                        key=lambda x: x["urn"],
                    )
                ]
            )

            aspect_map[Constants.ADD_OWNER_OPERATION] = owner_aspect

        if Constants.ADD_TERM_OPERATION in operation_map:
            term_aspect = mce_builder.make_glossary_terms_aspect_from_urn_list(
                sorted(set(operation_map[Constants.ADD_TERM_OPERATION]))
            )
            aspect_map[Constants.ADD_TERM_OPERATION] = term_aspect

        if Constants.ADD_DOC_LINK_OPERATION in operation_map:
            try:
                if len(
                    operation_map[Constants.ADD_DOC_LINK_OPERATION]
                ) == 1 and isinstance(
                    operation_map[Constants.ADD_DOC_LINK_OPERATION], list
                ):
                    docs_dict = cast(
                        List[Dict], operation_map[Constants.ADD_DOC_LINK_OPERATION]
                    )[0]
                    if "description" not in docs_dict or "link" not in docs_dict:
                        raise Exception(
                            "Documentation_link meta_mapping config needs a description key and a link key"
                        )

                    now = int(time.time() * 1000)  # milliseconds since epoch
                    institutional_memory_element = InstitutionalMemoryMetadataClass(
                        url=docs_dict["link"],
                        description=docs_dict["description"],
                        createStamp=AuditStampClass(
                            time=now, actor="urn:li:corpuser:ingestion"
                        ),
                    )

                    # create a new institutional memory aspect
                    institutional_memory_aspect = InstitutionalMemoryClass(
                        elements=[institutional_memory_element]
                    )

                    aspect_map[
                        Constants.ADD_DOC_LINK_OPERATION
                    ] = institutional_memory_aspect
                else:
                    raise Exception(
                        f"Expected 1 item of type list for the documentation_link meta_mapping config,"
                        f" received type of {type(operation_map[Constants.ADD_DOC_LINK_OPERATION])}"
                        f", and size of {len(operation_map[Constants.ADD_DOC_LINK_OPERATION])}."
                    )

            except Exception as e:
                logger.error(
                    f"Error while constructing aspect for documentation link and description : {e}"
                )

        return aspect_map

    def get_operation_value(
        self,
        operation_key: str,
        operation_type: str,
        operation_config: Dict,
        match: Match,
    ) -> Optional[Union[str, Dict, List[str], List[Dict]]]:
        if (
            operation_type == Constants.ADD_TAG_OPERATION
            and operation_config[Constants.TAG]
        ):
            tag = operation_config[Constants.TAG]
            tag = _insert_match_value(tag, _get_best_match(match, "tag"))

            if self.tag_prefix:
                tag = self.tag_prefix + tag
            return tag
        elif (
            operation_type == Constants.ADD_OWNER_OPERATION
            and operation_config[Constants.OWNER_TYPE]
        ):
            owner_id = _get_best_match(match, "owner")

            owner_ids: List[str] = [_id.strip() for _id in owner_id.split(",")]

            owner_category = (
                operation_config.get(Constants.OWNER_CATEGORY)
                or OwnershipTypeClass.DATAOWNER
            )
            owner_category, owner_category_urn = validate_ownership_type(owner_category)

            if self.strip_owner_email_id:
                owner_ids = [
                    self.sanitize_owner_ids(owner_id) for owner_id in owner_ids
                ]

            owner_type_mapping: Dict[str, OwnerType] = {
                Constants.USER_OWNER: OwnerType.USER,
                Constants.GROUP_OWNER: OwnerType.GROUP,
            }
            if operation_config[Constants.OWNER_TYPE] in owner_type_mapping:
                return _make_owner_category_list(
                    owner_ids=owner_ids,
                    owner_category=owner_category,
                    owner_category_urn=owner_category_urn,
                    owner_type=owner_type_mapping[
                        operation_config[Constants.OWNER_TYPE]
                    ],
                )

        elif (
            operation_type == Constants.ADD_TERM_OPERATION
            and operation_config[Constants.TERM]
        ):
            term = operation_config[Constants.TERM]
            term = _insert_match_value(term, _get_best_match(match, "term"))
            return mce_builder.make_term_urn(term)
        elif (
            operation_type == Constants.ADD_DOC_LINK_OPERATION
            and operation_config[Constants.DOC_LINK]
            and operation_config[Constants.DOC_DESCRIPTION]
        ):
            link = operation_config[Constants.DOC_LINK]
            link = _insert_match_value(link, _get_best_match(match, "link"))
            description = operation_config[Constants.DOC_DESCRIPTION]
            return {"link": link, "description": description}

        elif operation_type == Constants.ADD_TERMS_OPERATION:
            separator = operation_config.get(Constants.SEPARATOR, ",")
            captured_terms = match.group(0)
            return [
                mce_builder.make_term_urn(term.strip())
                for term in captured_terms.split(separator)
                if term.strip()
            ]
        return None

    def sanitize_owner_ids(self, owner_id: str) -> str:
        if owner_id.__contains__("@"):
            owner_id = owner_id[0 : owner_id.index("@")]
        return owner_id

    def get_match(self, match_clause: Any, raw_props_value: Any) -> Optional[Match]:
        # function to check if a match clause is satisfied to a value.
        if not any(
            isinstance(raw_props_value, t) for t in Constants.OPERAND_DATATYPE_SUPPORTED
        ) or not isinstance(raw_props_value, type(match_clause)):
            return None
        elif isinstance(raw_props_value, str):
            return re.match(match_clause, raw_props_value)
        else:
            return re.match(str(match_clause), str(raw_props_value))



import dataclasses
import os
import re
from typing import Any, ClassVar, Dict, List, Optional, Union, cast

import pydantic
from pydantic import Field, validator

from datahub.configuration import ConfigModel
from datahub.configuration.common import AllowDenyPattern
from datahub.configuration.source_common import (
    EnvConfigMixin,
    PlatformInstanceConfigMixin,
)
from datahub.configuration.validate_field_deprecation import pydantic_field_deprecated
from datahub.configuration.validate_field_removal import pydantic_removed_field
from datahub.ingestion.source.looker.looker_lib_wrapper import LookerAPIConfig
from datahub.ingestion.source.state.stale_entity_removal_handler import (
    StatefulStaleMetadataRemovalConfig,
)
from datahub.ingestion.source.state.stateful_ingestion_base import (
    StatefulIngestionConfigBase,
)


class NamingPattern(ConfigModel):
    ALLOWED_VARS: ClassVar[List[str]] = []
    REQUIRE_AT_LEAST_ONE_VAR: ClassVar[bool] = True

    pattern: str

    @classmethod
    def __get_validators__(cls):
        yield cls.pydantic_accept_raw_pattern
        yield cls.validate
        yield cls.pydantic_validate_pattern

    @classmethod
    def pydantic_accept_raw_pattern(cls, v):
        if isinstance(v, (NamingPattern, dict)):
            return v
        assert isinstance(v, str), "pattern must be a string"
        return {"pattern": v}

    @classmethod
    def pydantic_validate_pattern(cls, v):
        assert isinstance(v, NamingPattern)
        assert v.validate_pattern(cls.REQUIRE_AT_LEAST_ONE_VAR)
        return v

    @classmethod
    def allowed_docstring(cls) -> str:
        return f"Allowed variables are {cls.ALLOWED_VARS}"

    def validate_pattern(self, at_least_one: bool) -> bool:
        variables = re.findall("({[^}{]+})", self.pattern)

        variables = [v[1:-1] for v in variables]  # remove the {}

        for v in variables:
            if v not in self.ALLOWED_VARS:
                raise ValueError(
                    f"Failed to find {v} in allowed_variables {self.ALLOWED_VARS}"
                )
        if at_least_one and len(variables) == 0:
            raise ValueError(
                f"Failed to find any variable assigned to pattern {self.pattern}. Must have at least one. {self.allowed_docstring()}"
            )
        return True

    def replace_variables(self, values: Union[Dict[str, Optional[str]], object]) -> str:
        if not isinstance(values, dict):
            # Check that this is a dataclass instance (not a dataclass type).
            assert dataclasses.is_dataclass(values) and not isinstance(values, type)
            values = dataclasses.asdict(values)
        values = {k: v for k, v in values.items() if v is not None}
        return self.pattern.format(**values)


@dataclasses.dataclass
class NamingPatternMapping:
    platform: str
    env: str
    project: str
    model: str
    name: str


@dataclasses.dataclass
class ViewNamingPatternMapping(NamingPatternMapping):
    file_path: str
    folder_path: str


class LookerNamingPattern(NamingPattern):
    ALLOWED_VARS = [field.name for field in dataclasses.fields(NamingPatternMapping)]


class LookerViewNamingPattern(NamingPattern):
    ALLOWED_VARS = [
        field.name for field in dataclasses.fields(ViewNamingPatternMapping)
    ]


# TODO: deprecate browse_pattern configs
class LookerCommonConfig(EnvConfigMixin, PlatformInstanceConfigMixin):
    explore_naming_pattern: LookerNamingPattern = pydantic.Field(
        description=f"Pattern for providing dataset names to explores. {LookerNamingPattern.allowed_docstring()}",
        default=LookerNamingPattern(pattern="{model}.explore.{name}"),
    )
    explore_browse_pattern: LookerNamingPattern = pydantic.Field(
        description=f"Pattern for providing browse paths to explores. {LookerNamingPattern.allowed_docstring()}",
        default=LookerNamingPattern(pattern="/Explore/{model}"),
    )
    view_naming_pattern: LookerViewNamingPattern = Field(
        LookerViewNamingPattern(pattern="{project}.view.{name}"),
        description=f"Pattern for providing dataset names to views. {LookerViewNamingPattern.allowed_docstring()}",
    )
    view_browse_pattern: LookerViewNamingPattern = Field(
        LookerViewNamingPattern(pattern="/Develop/{project}/{folder_path}"),
        description=f"Pattern for providing browse paths to views. {LookerViewNamingPattern.allowed_docstring()}",
    )

    _deprecate_explore_browse_pattern = pydantic_field_deprecated(
        "explore_browse_pattern"
    )
    _deprecate_view_browse_pattern = pydantic_field_deprecated("view_browse_pattern")

    tag_measures_and_dimensions: bool = Field(
        True,
        description="When enabled, attaches tags to measures, dimensions and dimension groups to make them more "
        "discoverable. When disabled, adds this information to the description of the column.",
    )
    platform_name: str = Field(
        # TODO: This shouldn't be part of the config.
        "looker",
        description="Default platform name.",
        hidden_from_docs=True,
    )
    extract_column_level_lineage: bool = Field(
        True,
        description="When enabled, extracts column-level lineage from Views and Explores",
    )


class LookerDashboardSourceConfig(
    LookerAPIConfig,
    LookerCommonConfig,
    StatefulIngestionConfigBase,
    EnvConfigMixin,
):
    _removed_github_info = pydantic_removed_field("github_info")

    dashboard_pattern: AllowDenyPattern = Field(
        AllowDenyPattern.allow_all(),
        description="Patterns for selecting dashboard ids that are to be included",
    )
    chart_pattern: AllowDenyPattern = Field(
        AllowDenyPattern.allow_all(),
        description="Patterns for selecting chart ids that are to be included",
    )
    include_deleted: bool = Field(
        False,
        description="Whether to include deleted dashboards and looks.",
    )
    extract_owners: bool = Field(
        True,
        description="When enabled, extracts ownership from Looker directly. When disabled, ownership is left empty "
        "for dashboards and charts.",
    )
    strip_user_ids_from_email: bool = Field(
        False,
        description="When enabled, converts Looker user emails of the form name@domain.com to urn:li:corpuser:name "
        "when assigning ownership",
    )
    skip_personal_folders: bool = Field(
        False,
        description="Whether to skip ingestion of dashboards in personal folders. Setting this to True will only "
        "ingest dashboards in the Shared folder space.",
    )
    max_threads: int = Field(
        os.cpu_count() or 40,
        description="Max parallelism for Looker API calls. Defaults to cpuCount or 40",
    )
    external_base_url: Optional[str] = Field(
        None,
        description="Optional URL to use when constructing external URLs to Looker if the `base_url` is not the "
        "correct one to use. For example, `https://looker-public.company.com`. If not provided, "
        "the external base URL will default to `base_url`.",
    )
    extract_usage_history: bool = Field(
        True,
        description="Whether to ingest usage statistics for dashboards. Setting this to True will query looker system "
        "activity explores to fetch historical dashboard usage.",
    )
    # TODO - stateful ingestion to autodetect usage history interval
    extract_usage_history_for_interval: str = Field(
        "30 days",
        description="Used only if extract_usage_history is set to True. Interval to extract looker dashboard usage "
        "history for. See https://docs.looker.com/reference/filter-expressions#date_and_time.",
    )
    extract_embed_urls: bool = Field(
        True,
        description="Produce URLs used to render Looker Explores as Previews inside of DataHub UI. Embeds must be "
        "enabled inside of Looker to use this feature.",
    )
    stateful_ingestion: Optional[StatefulStaleMetadataRemovalConfig] = Field(
        default=None, description=""
    )
    extract_independent_looks: bool = Field(
        False,
        description="Extract looks which are not part of any Dashboard. To enable this flag the stateful_ingestion should also be enabled.",
    )
    emit_used_explores_only: bool = Field(
        True,
        description="When enabled, only explores that are used by a Dashboard/Look will be ingested.",
    )

    @validator("external_base_url", pre=True, always=True)
    def external_url_defaults_to_api_config_base_url(
        cls, v: Optional[str], *, values: Dict[str, Any], **kwargs: Dict[str, Any]
    ) -> Optional[str]:
        return v or values.get("base_url")

    @validator("extract_independent_looks", always=True)
    def stateful_ingestion_should_be_enabled(
        cls, v: Optional[bool], *, values: Dict[str, Any], **kwargs: Dict[str, Any]
    ) -> Optional[bool]:
        stateful_ingestion: StatefulStaleMetadataRemovalConfig = cast(
            StatefulStaleMetadataRemovalConfig, values.get("stateful_ingestion")
        )
        if v is True and (
            stateful_ingestion is None or stateful_ingestion.enabled is False
        ):
            raise ValueError("stateful_ingestion.enabled should be set to true")

        return v



# mypy: ignore-errors
# flake8: noqa

# This file is autogenerated by /metadata-ingestion/scripts/avro_codegen.py
# Do not modify manually!

# pylint: skip-file
# fmt: off
# isort: skip_file
from .....schema_classes import QueryLanguageClass
from .....schema_classes import QueryPropertiesClass
from .....schema_classes import QuerySourceClass
from .....schema_classes import QueryStatementClass
from .....schema_classes import QuerySubjectClass
from .....schema_classes import QuerySubjectsClass


QueryLanguage = QueryLanguageClass
QueryProperties = QueryPropertiesClass
QuerySource = QuerySourceClass
QueryStatement = QueryStatementClass
QuerySubject = QuerySubjectClass
QuerySubjects = QuerySubjectsClass

# fmt: on



from dataclasses import dataclass
from typing import Dict, Iterable, List, Optional, Tuple, Union

import feast.types
from feast import (
    BigQuerySource,
    Entity,
    FeatureStore,
    FeatureView,
    Field as FeastField,
    FileSource,
    KafkaSource,
    KinesisSource,
    OnDemandFeatureView,
    RequestSource,
    SnowflakeSource,
    ValueType,
)
from feast.data_source import DataSource
from pydantic import Field

import datahub.emitter.mce_builder as builder
from datahub.configuration.common import ConfigModel
from datahub.emitter.mce_builder import DEFAULT_ENV
from datahub.ingestion.api.common import PipelineContext
from datahub.ingestion.api.decorators import (
    SourceCapability,
    SupportStatus,
    capability,
    config_class,
    platform_name,
    support_status,
)
from datahub.ingestion.api.source import Source, SourceReport
from datahub.ingestion.api.workunit import MetadataWorkUnit
from datahub.metadata.com.linkedin.pegasus2avro.common import MLFeatureDataType
from datahub.metadata.com.linkedin.pegasus2avro.metadata.snapshot import (
    MLFeatureSnapshot,
    MLFeatureTableSnapshot,
    MLPrimaryKeySnapshot,
)
from datahub.metadata.com.linkedin.pegasus2avro.mxe import MetadataChangeEvent
from datahub.metadata.schema_classes import (
    BrowsePathsClass,
    MLFeaturePropertiesClass,
    MLFeatureTablePropertiesClass,
    MLPrimaryKeyPropertiesClass,
    StatusClass,
)

# FIXME: ValueType module cannot be used as a type
_field_type_mapping: Dict[Union[ValueType, feast.types.FeastType], str] = {
    ValueType.UNKNOWN: MLFeatureDataType.UNKNOWN,
    ValueType.BYTES: MLFeatureDataType.BYTE,
    ValueType.STRING: MLFeatureDataType.TEXT,
    ValueType.INT32: MLFeatureDataType.ORDINAL,
    ValueType.INT64: MLFeatureDataType.ORDINAL,
    ValueType.DOUBLE: MLFeatureDataType.CONTINUOUS,
    ValueType.FLOAT: MLFeatureDataType.CONTINUOUS,
    ValueType.BOOL: MLFeatureDataType.BINARY,
    ValueType.UNIX_TIMESTAMP: MLFeatureDataType.TIME,
    ValueType.BYTES_LIST: MLFeatureDataType.SEQUENCE,
    ValueType.STRING_LIST: MLFeatureDataType.SEQUENCE,
    ValueType.INT32_LIST: MLFeatureDataType.SEQUENCE,
    ValueType.INT64_LIST: MLFeatureDataType.SEQUENCE,
    ValueType.DOUBLE_LIST: MLFeatureDataType.SEQUENCE,
    ValueType.FLOAT_LIST: MLFeatureDataType.SEQUENCE,
    ValueType.BOOL_LIST: MLFeatureDataType.SEQUENCE,
    ValueType.UNIX_TIMESTAMP_LIST: MLFeatureDataType.SEQUENCE,
    ValueType.NULL: MLFeatureDataType.UNKNOWN,
    feast.types.Invalid: MLFeatureDataType.UNKNOWN,
    feast.types.Bytes: MLFeatureDataType.BYTE,
    feast.types.String: MLFeatureDataType.TEXT,
    feast.types.Int32: MLFeatureDataType.ORDINAL,
    feast.types.Int64: MLFeatureDataType.ORDINAL,
    feast.types.Float64: MLFeatureDataType.CONTINUOUS,
    feast.types.Float32: MLFeatureDataType.CONTINUOUS,
    feast.types.Bool: MLFeatureDataType.BINARY,
    feast.types.UnixTimestamp: MLFeatureDataType.TIME,
    feast.types.Array: MLFeatureDataType.SEQUENCE,  # type: ignore
    feast.types.Invalid: MLFeatureDataType.UNKNOWN,
}


class FeastRepositorySourceConfig(ConfigModel):
    path: str = Field(description="Path to Feast repository")
    fs_yaml_file: Optional[str] = Field(
        default=None,
        description="Path to the `feature_store.yaml` file used to configure the feature store",
    )
    environment: str = Field(
        default=DEFAULT_ENV, description="Environment to use when constructing URNs"
    )


@platform_name("Feast")
@config_class(FeastRepositorySourceConfig)
@support_status(SupportStatus.CERTIFIED)
@capability(SourceCapability.LINEAGE_COARSE, "Enabled by default")
@dataclass
class FeastRepositorySource(Source):
    """
    This plugin extracts:

    - Entities as [`MLPrimaryKey`](https://datahubproject.io/docs/graphql/objects#mlprimarykey)
    - Fields as [`MLFeature`](https://datahubproject.io/docs/graphql/objects#mlfeature)
    - Feature views and on-demand feature views as [`MLFeatureTable`](https://datahubproject.io/docs/graphql/objects#mlfeaturetable)
    - Batch and stream source details as [`Dataset`](https://datahubproject.io/docs/graphql/objects#dataset)
    - Column types associated with each entity and feature
    """

    platform = "feast"
    source_config: FeastRepositorySourceConfig
    report: SourceReport
    feature_store: FeatureStore

    def __init__(self, config: FeastRepositorySourceConfig, ctx: PipelineContext):
        super().__init__(ctx)
        self.source_config = config
        self.report = SourceReport()
        self.feature_store = FeatureStore(
            repo_path=self.source_config.path,
            fs_yaml_file=self.source_config.fs_yaml_file,
        )

    def _get_field_type(
        self, field_type: Union[ValueType, feast.types.FeastType], parent_name: str
    ) -> str:
        """
        Maps types encountered in Feast to corresponding schema types.
        """

        ml_feature_data_type = _field_type_mapping.get(field_type)

        if ml_feature_data_type is None:
            self.report.report_warning(
                parent_name, f"unable to map type {field_type} to metadata schema"
            )

            ml_feature_data_type = MLFeatureDataType.UNKNOWN

        return ml_feature_data_type

    def _get_data_source_details(self, source: DataSource) -> Tuple[str, str]:
        """
        Get Feast batch/stream source platform and name.
        """

        platform = "unknown"
        name = "unknown"

        if isinstance(source, FileSource):
            platform = "file"
            name = source.path.replace("://", ".").replace("/", ".")
        elif isinstance(source, BigQuerySource):
            platform = "bigquery"
            name = source.table
        elif isinstance(source, KafkaSource):
            platform = "kafka"
            name = source.kafka_options.topic
        elif isinstance(source, KinesisSource):
            platform = "kinesis"
            name = (
                f"{source.kinesis_options.region}:{source.kinesis_options.stream_name}"
            )
        elif isinstance(source, RequestSource):
            platform = "request"
            name = source.name
        elif isinstance(source, SnowflakeSource):
            platform = "snowflake"
            name = source.table

        return platform, name

    def _get_data_sources(self, feature_view: FeatureView) -> List[str]:
        """
        Get data source URN list.
        """

        sources = []

        if feature_view.batch_source is not None:
            batch_source_platform, batch_source_name = self._get_data_source_details(
                feature_view.batch_source
            )
            sources.append(
                builder.make_dataset_urn(
                    batch_source_platform,
                    batch_source_name,
                    self.source_config.environment,
                )
            )

        if feature_view.stream_source is not None:
            stream_source_platform, stream_source_name = self._get_data_source_details(
                feature_view.stream_source
            )
            sources.append(
                builder.make_dataset_urn(
                    stream_source_platform,
                    stream_source_name,
                    self.source_config.environment,
                )
            )

        return sources

    def _get_entity_workunit(
        self, feature_view: FeatureView, entity: Entity
    ) -> MetadataWorkUnit:
        """
        Generate an MLPrimaryKey work unit for a Feast entity.
        """

        feature_view_name = f"{self.feature_store.project}.{feature_view.name}"

        entity_snapshot = MLPrimaryKeySnapshot(
            urn=builder.make_ml_primary_key_urn(feature_view_name, entity.name),
            aspects=[StatusClass(removed=False)],
        )

        entity_snapshot.aspects.append(
            MLPrimaryKeyPropertiesClass(
                description=entity.description,
                dataType=self._get_field_type(entity.value_type, entity.name),
                sources=self._get_data_sources(feature_view),
            )
        )

        mce = MetadataChangeEvent(proposedSnapshot=entity_snapshot)

        return MetadataWorkUnit(id=entity.name, mce=mce)

    def _get_feature_workunit(
        self,
        # FIXME: FeatureView and OnDemandFeatureView cannot be used as a type
        feature_view: Union[FeatureView, OnDemandFeatureView],
        field: FeastField,
    ) -> MetadataWorkUnit:
        """
        Generate an MLFeature work unit for a Feast feature.
        """
        feature_view_name = f"{self.feature_store.project}.{feature_view.name}"

        feature_snapshot = MLFeatureSnapshot(
            urn=builder.make_ml_feature_urn(feature_view_name, field.name),
            aspects=[StatusClass(removed=False)],
        )

        feature_sources = []
        if isinstance(feature_view, FeatureView):
            feature_sources = self._get_data_sources(feature_view)
        elif isinstance(feature_view, OnDemandFeatureView):
            if feature_view.source_request_sources is not None:
                for request_source in feature_view.source_request_sources.values():
                    source_platform, source_name = self._get_data_source_details(
                        request_source
                    )

                    feature_sources.append(
                        builder.make_dataset_urn(
                            source_platform,
                            source_name,
                            self.source_config.environment,
                        )
                    )

            if feature_view.source_feature_view_projections is not None:
                for (
                    feature_view_projection
                ) in feature_view.source_feature_view_projections.values():
                    feature_view_source = self.feature_store.get_feature_view(
                        feature_view_projection.name
                    )

                    feature_sources.extend(self._get_data_sources(feature_view_source))

        feature_snapshot.aspects.append(
            MLFeaturePropertiesClass(
                description=field.tags.get("description"),
                dataType=self._get_field_type(field.dtype, field.name),
                sources=feature_sources,
            )
        )

        mce = MetadataChangeEvent(proposedSnapshot=feature_snapshot)

        return MetadataWorkUnit(id=field.name, mce=mce)

    def _get_feature_view_workunit(self, feature_view: FeatureView) -> MetadataWorkUnit:
        """
        Generate an MLFeatureTable work unit for a Feast feature view.
        """

        feature_view_name = f"{self.feature_store.project}.{feature_view.name}"

        feature_view_snapshot = MLFeatureTableSnapshot(
            urn=builder.make_ml_feature_table_urn("feast", feature_view_name),
            aspects=[
                BrowsePathsClass(paths=[f"/feast/{self.feature_store.project}"]),
                StatusClass(removed=False),
            ],
        )

        feature_view_snapshot.aspects.append(
            MLFeatureTablePropertiesClass(
                mlFeatures=[
                    builder.make_ml_feature_urn(
                        feature_view_name,
                        feature.name,
                    )
                    for feature in feature_view.features
                ],
                mlPrimaryKeys=[
                    builder.make_ml_primary_key_urn(feature_view_name, entity_name)
                    for entity_name in feature_view.entities
                ],
            )
        )

        mce = MetadataChangeEvent(proposedSnapshot=feature_view_snapshot)

        return MetadataWorkUnit(id=feature_view_name, mce=mce)

    def _get_on_demand_feature_view_workunit(
        self, on_demand_feature_view: OnDemandFeatureView
    ) -> MetadataWorkUnit:
        """
        Generate an MLFeatureTable work unit for a Feast on-demand feature view.
        """

        on_demand_feature_view_name = (
            f"{self.feature_store.project}.{on_demand_feature_view.name}"
        )

        on_demand_feature_view_snapshot = MLFeatureTableSnapshot(
            urn=builder.make_ml_feature_table_urn("feast", on_demand_feature_view_name),
            aspects=[
                BrowsePathsClass(paths=[f"/feast/{self.feature_store.project}"]),
                StatusClass(removed=False),
            ],
        )

        on_demand_feature_view_snapshot.aspects.append(
            MLFeatureTablePropertiesClass(
                mlFeatures=[
                    builder.make_ml_feature_urn(
                        on_demand_feature_view_name,
                        feature.name,
                    )
                    for feature in on_demand_feature_view.features
                ],
                mlPrimaryKeys=[],
            )
        )

        mce = MetadataChangeEvent(proposedSnapshot=on_demand_feature_view_snapshot)

        return MetadataWorkUnit(id=on_demand_feature_view_name, mce=mce)

    @classmethod
    def create(cls, config_dict, ctx):
        config = FeastRepositorySourceConfig.parse_obj(config_dict)
        return cls(config, ctx)

    def get_workunits_internal(self) -> Iterable[MetadataWorkUnit]:
        for feature_view in self.feature_store.list_feature_views():
            for entity_name in feature_view.entities:
                entity = self.feature_store.get_entity(entity_name)
                yield self._get_entity_workunit(feature_view, entity)

            for field in feature_view.features:
                yield self._get_feature_workunit(feature_view, field)

            yield self._get_feature_view_workunit(feature_view)

        for on_demand_feature_view in self.feature_store.list_on_demand_feature_views():
            for feature in on_demand_feature_view.features:
                yield self._get_feature_workunit(on_demand_feature_view, feature)

            yield self._get_on_demand_feature_view_workunit(on_demand_feature_view)

    def get_report(self) -> SourceReport:
        return self.report



from dataclasses import dataclass, field
from datetime import datetime
from typing import Dict, List, MutableSet, Optional

from datahub.ingestion.api.report import Report
from datahub.ingestion.glossary.classification_mixin import ClassificationReportMixin
from datahub.ingestion.source.snowflake.constants import SnowflakeEdition
from datahub.ingestion.source.sql.sql_generic_profiler import ProfilingSqlReport
from datahub.ingestion.source.state.stateful_ingestion_base import (
    StatefulIngestionReport,
)
from datahub.ingestion.source_report.ingestion_stage import IngestionStageReport
from datahub.ingestion.source_report.time_window import BaseTimeWindowReport
from datahub.sql_parsing.sql_parsing_aggregator import SqlAggregatorReport
from datahub.utilities.perf_timer import PerfTimer


@dataclass
class SnowflakeUsageAggregationReport(Report):
    query_secs: float = -1
    query_row_count: int = -1
    result_fetch_timer: PerfTimer = field(default_factory=PerfTimer)
    result_skip_timer: PerfTimer = field(default_factory=PerfTimer)
    result_map_timer: PerfTimer = field(default_factory=PerfTimer)
    users_map_timer: PerfTimer = field(default_factory=PerfTimer)
    queries_map_timer: PerfTimer = field(default_factory=PerfTimer)
    fields_map_timer: PerfTimer = field(default_factory=PerfTimer)


@dataclass
class SnowflakeUsageReport:
    min_access_history_time: Optional[datetime] = None
    max_access_history_time: Optional[datetime] = None
    access_history_range_query_secs: float = -1
    access_history_query_secs: float = -1

    rows_processed: int = 0
    rows_missing_query_text: int = 0
    rows_zero_base_objects_accessed: int = 0
    rows_zero_direct_objects_accessed: int = 0
    rows_missing_email: int = 0
    rows_parsing_error: int = 0

    usage_start_time: Optional[datetime] = None
    usage_end_time: Optional[datetime] = None
    stateful_usage_ingestion_enabled: bool = False

    usage_aggregation: SnowflakeUsageAggregationReport = field(
        default_factory=SnowflakeUsageAggregationReport
    )


@dataclass
class SnowflakeReport(ProfilingSqlReport, BaseTimeWindowReport):
    num_table_to_table_edges_scanned: int = 0
    num_table_to_view_edges_scanned: int = 0
    num_view_to_table_edges_scanned: int = 0
    num_external_table_edges_scanned: int = 0
    ignore_start_time_lineage: Optional[bool] = None
    upstream_lineage_in_report: Optional[bool] = None
    upstream_lineage: Dict[str, List[str]] = field(default_factory=dict)

    lineage_start_time: Optional[datetime] = None
    lineage_end_time: Optional[datetime] = None
    stateful_lineage_ingestion_enabled: bool = False

    cleaned_account_id: str = ""
    run_ingestion: bool = False

    # https://community.snowflake.com/s/topic/0TO0Z000000Unu5WAC/releases
    saas_version: Optional[str] = None
    default_warehouse: Optional[str] = None
    default_db: Optional[str] = None
    default_schema: Optional[str] = None
    role: str = ""

    profile_if_updated_since: Optional[datetime] = None
    profile_candidates: Dict[str, List[str]] = field(default_factory=dict)

    # lineage/usage v2
    sql_aggregator: Optional[SqlAggregatorReport] = None


@dataclass
class SnowflakeV2Report(
    SnowflakeReport,
    SnowflakeUsageReport,
    StatefulIngestionReport,
    ClassificationReportMixin,
    IngestionStageReport,
):
    account_locator: Optional[str] = None
    region: Optional[str] = None

    schemas_scanned: int = 0
    databases_scanned: int = 0
    tags_scanned: int = 0

    include_usage_stats: bool = False
    include_operational_stats: bool = False
    include_technical_schema: bool = False
    include_column_lineage: bool = False

    table_lineage_query_secs: float = -1
    external_lineage_queries_secs: float = -1
    num_tables_with_known_upstreams: int = 0
    num_upstream_lineage_edge_parsing_failed: int = 0

    # Reports how many times we reset in-memory `functools.lru_cache` caches of data,
    # which occurs when we occur a different database / schema.
    # Should not be more than the number of databases / schemas scanned.
    # Maps (function name) -> (stat_name) -> (stat_value)
    lru_cache_info: Dict[str, Dict[str, int]] = field(default_factory=dict)

    # These will be non-zero if snowflake information_schema queries fail with error -
    # "Information schema query returned too much data. Please repeat query with more selective predicates.""
    # This will result in overall increase in time complexity
    num_get_tables_for_schema_queries: int = 0
    num_get_views_for_schema_queries: int = 0
    num_get_columns_for_table_queries: int = 0

    # these will be non-zero if the user choses to enable the extract_tags = "with_lineage" option, which requires
    # individual queries per object (database, schema, table) and an extra query per table to get the tags on the columns.
    num_get_tags_for_object_queries: int = 0
    num_get_tags_on_columns_for_table_queries: int = 0

    rows_zero_objects_modified: int = 0

    _processed_tags: MutableSet[str] = field(default_factory=set)
    _scanned_tags: MutableSet[str] = field(default_factory=set)

    edition: Optional[SnowflakeEdition] = None

    def report_entity_scanned(self, name: str, ent_type: str = "table") -> None:
        """
        Entity could be a view or a table or a schema or a database
        """
        if ent_type == "table":
            self.tables_scanned += 1
        elif ent_type == "view":
            self.views_scanned += 1
        elif ent_type == "schema":
            self.schemas_scanned += 1
        elif ent_type == "database":
            self.databases_scanned += 1
        elif ent_type == "tag":
            # the same tag can be assigned to multiple objects, so we need
            # some extra logic account for each tag only once.
            if self._is_tag_scanned(name):
                return
            self._scanned_tags.add(name)
            self.tags_scanned += 1
        else:
            raise KeyError(f"Unknown entity {ent_type}.")

    def is_tag_processed(self, tag_name: str) -> bool:
        return tag_name in self._processed_tags

    def _is_tag_scanned(self, tag_name: str) -> bool:
        return tag_name in self._scanned_tags

    def report_tag_processed(self, tag_name: str) -> None:
        self._processed_tags.add(tag_name)

    def set_ingestion_stage(self, database: str, stage: str) -> None:
        self.report_ingestion_stage_start(f"{database}: {stage}")



import copy
import re
from typing import Any, Dict, Optional, cast

from datahub.configuration.common import ConfigModel
from datahub.emitter.mce_builder import Aspect
from datahub.ingestion.api.common import PipelineContext
from datahub.ingestion.transformer.dataset_transformer import (
    DatasetPropertiesTransformer,
)
from datahub.metadata.schema_classes import DatasetPropertiesClass


class ReplaceExternalUrlConfig(ConfigModel):
    input_pattern: str
    replacement: str


class ReplaceExternalUrl(DatasetPropertiesTransformer):
    """Transformer that clean the ownership URN."""

    ctx: PipelineContext
    config: ReplaceExternalUrlConfig

    def __init__(
        self,
        config: ReplaceExternalUrlConfig,
        ctx: PipelineContext,
        **resolver_args: Dict[str, Any],
    ):
        super().__init__()
        self.ctx = ctx
        self.config = config
        self.resolver_args = resolver_args

    @classmethod
    def create(cls, config_dict: dict, ctx: PipelineContext) -> "ReplaceExternalUrl":
        config = ReplaceExternalUrlConfig.parse_obj(config_dict)
        return cls(config, ctx)

    def transform_aspect(
        self, entity_urn: str, aspect_name: str, aspect: Optional[Aspect]
    ) -> Optional[Aspect]:
        in_dataset_properties_aspect: DatasetPropertiesClass = cast(
            DatasetPropertiesClass, aspect
        )

        if (
            not hasattr(in_dataset_properties_aspect, "externalUrl")
            or not in_dataset_properties_aspect.externalUrl
        ):
            return cast(Aspect, in_dataset_properties_aspect)
        else:
            out_dataset_properties_aspect: DatasetPropertiesClass = copy.deepcopy(
                in_dataset_properties_aspect
            )

            pattern = re.compile(self.config.input_pattern)
            replacement = self.config.replacement

            out_dataset_properties_aspect.externalUrl = re.sub(
                pattern, replacement, in_dataset_properties_aspect.externalUrl
            )

            return cast(Aspect, out_dataset_properties_aspect)
